#!/usr/bin/env python3
# coding: utf-8
""""""
import os
import csv

from io import StringIO
from zipfile import ZipFile
from geojson import Feature, Point, FeatureCollection
from openpyxl import Workbook
from openpyxl.utils import get_column_letter

from ogt.thirdparty import bng_to_latlon

from . import utils
from .ogt_group import OGTGroup
from .ogt_cell import OGTCell
from . import ogt_validate
from . import ags4

from . import FORMATS


class OGTDocumentOptions:
    """"""
    def __init__(self):
        self.minify = False
        """Option whether to minify output Json only"""

        self.extended = True
        """Option to `extended` output with heading/group descriptions, etc"""

        self.include_stats = False
        """Stats such as groups, row count etc"""

        self.include_source = False
        """Includes `source`  and `source_cells` in output """

    def __repr__(self):
        return "<OGTDocOpts extended=%s, mini=%s>" % (self.extended, self.minify)


class OGTDocument:
    """:class:`~ogt.ogt_doc.OGTDocument`  is primary container of data (:class:`~ogt.ogt_group.OGTGroup`).



    .. code-block:: python

        from ogt import ogt_doc

        doc = ogt_doc.OGTDocument()
        err = doc.load_ags4_file("/path/to/my.ags")
        if err:
            print err
        else:
            # print the groups index
            print doc.groups_index()

            # Headings in the SAMP group
            print doc.group("SAMP").headings()

            # Return a list of units used in the document
            print doc.units()

    """

    class SORT:
        source = 0
        """Same order as source file"""

        a2z = 1
        """Alphabatical sort"""

        recommended = 2
        """Prefered sort see :func:`groups_sort`"""

    def __init__(self, sort_mode=SORT.recommended):


        self.source_refs = []
        """Source refs ef the file_name or url, nick etc """

        self.source_strings = {}
        """The original source files contents as string
            TODO: need to sort out utf8 vs ascii vs windows characters
        """

        self._groups_index_src = []
        """A list of groups in imported order"""

        self._groups_index_curr = []
        self._sort_mode = sort_mode

        self.groups_dict = {}
        """A `dict` of group code to :class:`~ogt.ogt_group.OGTGroup` instances

        :type: dict[str] => OGTGroup
        """

        self._errors_cache = None
        """A `dict` of row/col indexes with errors"""

        self.opts = OGTDocumentOptions()
        """Set default options :class:`~ogt.ogt_doc.OGTDocumentOptions` """

        self.sig_changed = None
        """A pointer to a `pyqtSignal` (only used with gui) see :meth:`emit` """

    def emit(self):
        """Emits the :attr:`sig_changed` if set"""
        print("ogtDoc.emit()", self)
        if self.sig_changed is None:
            return
        self.sig_changed.emit()

    def concat_char(self, default="+"):
        """
        :return: conantenation character (see :ref:`ags4_rule_11b`)
        :type: str
        """
        grp = self.group_from_code("TRAN")
        if len(grp.data) == 1:
            cell = grp.data[0].get("TRAN_RCON")
            if cell:
                return cell.value
        return default

    def add_get_group(self, cell_or_val, src_ref=None):
        """Adds (if not exist) or gets (if exists) a :class:`~ogt.ogt_group.OGTGroup` instance in this document

        :param grp: The group object to add
        :type grp: ~ogt.ogt_group.OGTGroup
        :return: An `Error` message is group exists, else `None`
        """

        gcell = cell_or_val if isinstance(cell_or_val, OGTCell) else OGTCell(raw=cell_or_val)

        # first check group not already there
        group_code = gcell.value

        ogtGroup = self.groups_dict.get(group_code)
        if ogtGroup is None:
            # not exist so add
            ogtGroup = OGTGroup(self, gcell, src_ref=src_ref)
            self.groups_dict[group_code] = ogtGroup

            self._groups_index_src.append(group_code)
            self._groups_index_curr.append(group_code)

        return ogtGroup

    def add_required_groups(self):
        for grp_code in GROUPS_REQUIRED:
            cell = OGTCell(raw=grp_code)
            grp = self.add_get_group(cell, src_ref="added")

            hcodes = GROUP_HEADINGS_DEFAULT.get(grp.group_code)
            cells = {}
            if hcodes:
                for hc in hcodes:
                    grpdd = ags4.DD.group(grp.group_code)
                    idx = grpdd['headings_index'].index(hc)
                    rec = grpdd['headings'][idx]
                    grp.add_heading(rec)

                    if len(grp.data) == 0:
                        cell = OGTCell(raw="")
                        cells[hc] = cell
                grp.add_data_row(cells)

        self.update_sort()

        self.emit()

    def set_sort_mode(self, sort_mode):
        """
        :param sort_mode: The index of the group
        :type sort_mode:
        Return a list of group_codes in preferred order (see :func:`~ogt.ogt_group.groups_sort`)



        """
        self._sort_mode = sort_mode
        self.update_sort()

    def update_sort(self):
        if self._sort_mode == self.SORT.a2z:
            self._groups_index_curr = sorted(self._groups_index_src)

        elif self._sort_mode == self.SORT.source:
            self._groups_index_curr = self._groups_index_src
        else:
            self._groups_index_curr = groups_sort(self._groups_index_src)

    def group_from_index(self, idx):
        """
        :param idx: The index of the group
        :type idx: int
        :rtype: OGTGroup
        :return: Groups codes index
        """
        return self.groups_dict[self._groups_index_curr[idx]]

    @property
    def groups_list(self):
        """
        :type: list[OGTGroup]
        :return: A `list` of :class:`OgtGroup` objects
        """
        return [self.groups_dict[gc] for gc in self._groups_index_curr]

    @property
    def groups_index(self):
        """
        :rtype: list[str]
        :return: Groups codes index
        """
        return self._groups_index_curr

    @property
    def groups_count(self):
        """
        :rtype: int
        :return: Returns no of groups in the document
        """
        return len(self._groups_index_src)

    def group_from_code(self, group_code):
        """
        :param group_code: Four character group code
        :type group_code: str
        :return: An instance of :class:`~ogt.ogt_group.OGTGroup` if exists, else `None`
        """
        return self.groups_dict.get(group_code)

    def proj(self):
        """Shortcut to `PROJ` group object

        :return: An instance of :class:`~ogt.ogt_group.OGTGroup` if exists, else `None`
        """
        return self.group_from_code("PROJ")

    def proj_dict(self):
        """Shortcut to `PROJ` group data

        :return: A dict with data if exists, else `None`
        """
        grpOb = self.group_from_code("PROJ")
        if not grpOb:
            return None
        if grpOb.data_rows_count > 0:  # should always be one on project row
            return grpOb.data_row_dict(0)
        return None

    def units(self):
        """Shortcut to `UNIT` group

        :rtype: tuple
        :return:
            - An instance of :class:`~ogt.ogt_group.OGTGroup` if exists, else `None`
            - `bool` = `True` if group found in document, else false
        """
        return self.group_from_code("UNIT")

    def data_types(self):
        """Shortcut to `TYPE` group

        :rtype: tuple
        :return:
            - An instance of :class:`~ogt.ogt_group.OGTGroup` if exists, else `None`
            - `bool` = `True` if group found in document, else false
        """
        return self.group_from_code("TYPE")

    def frsfabbrs(self, head_code):
        """"""
        lst = []
        grp = self.group_from_code("ABBR")
        if grp:
            for cellsDic in grp.data:
                cellH = cellsDic.get("ABBR_HDNG")
                if cellH and cellH.value == head_code:
                    dic = dict(code=cellsDic.get("ABBR_CODE").value,
                               description=cellsDic.get("ABBR_DESC").value, custom=True,
                               list="CUSTOM")
                    lst.append(dic)
        ags_abbrs = ags4.DD.picklist_vals(head_code)
        if ags_abbrs:
            lst.extend(ags_abbrs)

    def add_abbr(self, head_code, rec):
        """Add abbreviation row"""
        # TODO inset index

        existing_abbrs = self.abbrs(head_code, as_dict=True)
        if rec['code'] not in existing_abbrs:

            row = {}
            row['ABBR_HDNG'] = OGTCell(raw=head_code)
            row['ABBR_CODE'] = OGTCell(raw=rec['code'])
            row['ABBR_DESC'] = OGTCell(raw=rec['description'])
            row['ABBR_LIST'] = OGTCell(raw=rec['list'])

            grp = self.group_from_code("ABBR")

            # insert at end of heading block, otherwise end
            found_group = False
            for ridx, rec in enumerate(grp.data_rows_dict()):

                if rec['ABBR_HDNG'] == head_code:
                    found_group = True

                if rec['ABBR_HDNG'] != head_code and found_group:
                    grp.insert_data_row(ridx, row)
                    self.emit()
                    return

            grp.add_data_row(row)

    def abbrs(self, head_code, as_dict=False, code_list=False):
        """Abbtrviatiosn for heading"""
        lookup = {}
        code_idx = []

        # add abbrs from ags
        agscodes = ags4.DD.abbrs(head_code)
        if agscodes:
            for rec in agscodes:
                lookup[rec['code']] = rec
                code_idx.append(rec['code'])

        # add abbrs from this file
        grp = self.group_from_code("ABBR")
        if grp:
            for cellsDic in grp.data:
                cellH = cellsDic.get("ABBR_HDNG")
                if cellH and cellH.value == head_code:
                    code = cellsDic.get("ABBR_CODE").value
                    if code in lookup:
                        # already in ags4
                        continue
                    dic = dict(code=code, description=cellsDic.get("ABBR_DESC").value,
                               list=cellsDic.get("ABBR_LIST").value)
                    lookup[dic['code']] = dic
                    code_idx.append(dic['code'])

        if as_dict:
            return lookup
        if code_list:
            return sorted(code_idx)
        return [lookup[code] for code in sorted(code_idx)]

    # def errors(self, lidx=None, cidx=None):
    #     """"""
    #     if lidx is not None:
    #         recs = self.error_cells.get(lidx)
    #         if recs is None:
    #             return None
    #         if cidx is None:
    #             return recs
    #         return recs.get(cidx)
    #     return None

    def errors_list(self):
        """"""
        if self._errors_cache is None:
            self._update_errors_cache()
        return self._errors_cache
        lst = []
        for ridx, row in enumerate(self.cells):
            for cidx, cell in enumerate(row):
                lst.extend(cell.errors)

        return lst

        lst = []
        for lidx in sorted(self.error_cells.keys()):
            for cidx in sorted(self.error_cells[lidx].keys()):
                lst.extend(self.error_cells[lidx][cidx])
        return lst

    def _update_errors_cache(self):
        """"""
        self._errors_cache = []
        for grp in self.groups_dict.values():
            self._errors_cache.extend(grp.errors_list())

    @property
    def errors_warnings_count(self):
        """"""
        if self._errors_cache is None:
            self._update_errors_cache()

        return len(self._errors_cache)

    def clear_errors(self):
        """Clears all errors, eg before revalidation"""
        self._errors_cache = None
        for grp in self.groups_dict.values():
            grp.clear_errors()

    def validate(self):
        """WIP validate doc

        """
        # first clear all existing errors
        self.clear_errors()

        for grp in self.groups_dict.values():
            grp.validate()

        # validate each group
        for grp in self.groups_dict.values():
            grp.validate()

    def write(self, ext="json", beside=False, file_path=None,
              to_zip=False, overwrite=False):
        """Write out the data to file in the selected format

        .. Note:: **Note**

            - Either **`beside=True`** or a **`file_path`** is required, otherwise and error occurs
            - If both are provided, and error is returned

        :param ext: The file format, see :data:`~ogt.__init__.FORMATS`
        :type ext: str

        :type beside: bool
        :param beside: Save the output file alongside the original with extention appended, eg

             - Source = `/path/to/myproject.ags`
             - Output = `/path/to/myproject.ags.json`

        :param file_path: Relative or absolute path to write to including extention
        :type file_path: str

        :param include_source: If `True`, the original ags source is also included.
        :type include_source: bool

        :param to_zip: If `True`, the original and converted file are packaged in a zip
        :type to_zip: bool

        :param overwrite: If `True`, the target file is overwritten, otherwise an error is returned
        :type overwrite: bool

        :return: A tuple with

                - A `Message` string  if no errors, else `None`
                - Any `Error` that occured, otherwise `None`

        """
        self.stats()['site_geometry']

        # Do some validations
        if ext not in FORMATS:
            return None, "Error: Invalid format specified - `%s` % ext. Use %s" % ",".join(FORMATS)

        if not beside and file_path is None:
            return None, "Error: need an output, either -b or -w"

        if beside and file_path is not None:
            return None, "Error: conflict in options, either -b or -w, not BOTH"

        # make target filename's
        base_name = os.path.basename(self.source_file_path)
        target_file_path = None
        if beside:
            # File is beside the original
            if to_zip:
                target_file_path = self.source_file_path + ".zip"

            else:
                target_file_path = self.source_file_path + ".%s" % ext

        else:
            # file is from argument
            target_file_path = file_path
            base_name = os.path.basename(file_path)
            if len(base_name) == 0:
                # directory given only
                return None, "Error: Invalid file name `%s`" % target_file_path

            parts = base_name.split(".")
            if len(parts) == 1:
                # no extention
                return None, "Error: Invalid file name `%s`" % target_file_path

            # Check the extention is what we expect
            gext = parts[-1]
            if not to_zip and gext != ext:
                return None, "Error: Conflict in file name extention, expected '%s' `%s`" % (ext, target_file_path)
            elif to_zip and gext != "zip":
                # extentions mismatched eg json != yaml
                return None, "Error: Conflict in file name extention expected 'zip' `%s`" % target_file_path

        # TODO warn if not overwrite
        if not overwrite:
            if os.path.exists(target_file_path):
                return None, "Error: Target file exists - `%s` " % target_file_path

        # Excel get own function
        if ext == "xlsx":
            return self.write_excel(target_file_path)

        # convert the file to target format string blob
        blob = None
        err = None
        if ext in ["js", "json"]:
            blob, err = self.to_json()

        elif ext == "geojson":
            blob, err = self.to_geojson()

        elif ext == "yaml":
            blob, err = self.to_yaml()

        elif ext in ["ags", "ags4"]:
            blob, err = self.to_ags4()

        else:
            return None, "Error: No valid output format specified - `%s` % ext"

        if err:
            return None, err

        if to_zip:
            # create zip
            try:
                zipee = ZipFile(target_file_path, mode="w")

                # add source file
                zipee.writestr(base_name, self.source_file_path)

                # add converted file
                zipee.writestr("%s.%s" % (base_name, ext), blob)

                # write out and done
                zipee.close()

                siz = utils.file_size(target_file_path, human=True)
                return "Wrote: %s `%s`" % (siz, target_file_path), None

            except Exception as error:
                return None, "Error: %s" % str(error)
        else:
            try:
                with open(target_file_path, "w") as file:
                    file.write(blob)
                    file.close()
                siz = utils.file_size(target_file_path, human=True)
                return "Wrote: %s `%s`" % (siz, target_file_path), None

            except Exception as error:
                return None, "Error: %s" % str(error)

        return None, "Error: OOPS unexpected error"

    def to_ags4(self):
        """

        :return: The document as an ags string
        :rtype: str or None, err
        """
        out = StringIO()
        writer = csv.writer(out, delimiter=',', lineterminator='\r\n',
                            quotechar='"', quoting=csv.QUOTE_ALL)

        for group_code in self.groups_index:
            grp = self.groups_dict[group_code]

            # write "GROUP"
            writer.writerow([ags4.DESCRIPTOR.group, group_code])

            # write headings
            lst = [ags4.DESCRIPTOR.heading]
            lst.extend(grp.headings_index)
            writer.writerow(lst)

            # write units
            lst = [ags4.DESCRIPTOR.unit]
            lst.extend(grp.units_list())
            writer.writerow(lst)

            # write types
            lst = [ags4.DESCRIPTOR.data_type]
            lst.extend(grp.data_types_list())
            writer.writerow(lst)

            # write data
            for dic in grp.data:
                lst = [ags4.DESCRIPTOR.data]
                for hcode in grp.headings_index:
                    v = ""
                    cell = dic.get(hcode)
                    if cell:
                        v = cell.value
                    lst.append(v)
                writer.writerow(lst)

            writer.writerow([])

        return out.getvalue(), None

    def to_grid(self):
        grid = []
        max_col = 0

        def to_dic(lst):
            dic = {}
            for idx, v in enumerate(lst):
                dic[idx] = v
            return dic

        for group_code in self.groups_index:
            grp = self.groups_dict[group_code]

            # write "GROUP"
            grid.append(to_dic([ags4.DESCRIPTOR.group, grp.group_code_cell]))

            # write headings
            lst = [ags4.DESCRIPTOR.heading]
            for hh in grp.headings_list:
                lst.append(hh.head_cell)

            grid.append(to_dic(lst))
            if len(lst) > max_col:
                max_col = len(lst)

            # write units
            lst = [ags4.DESCRIPTOR.unit]
            for hh in grp.headings_list:
                lst.append(hh.unit_cell)
            grid.append(to_dic(lst))

            # write types
            lst = [ags4.DESCRIPTOR.data_type]
            for hh in grp.headings_list:
                lst.append(hh.data_type_cell)
            grid.append(to_dic(lst))

            # write data
            for dic in grp.data:
                lst = [ags4.DESCRIPTOR.data]
                for hcode in grp.headings_index:
                    cell = dic.get(hcode)
                    lst.append(cell)
                grid.append(to_dic(lst))

            grid.append(to_dic([]))

        return grid, max_col

    def to_dict(self):
        """
        :rtype: dict
        :return: The data in a dict, see :attr:`OGTDocument.opts` for serialisation options
        """
        # base dict to return
        dic = dict(file_name=self.source_file_path, version="ags4", groups={})

        # loop groups and add struct based on edit_mode
        for k, g in self.groups_dict.items():
            dic['groups'][k] = g.to_dict()

        # include source raw source
        if self.opts.include_source:
            pass

            # TODO
            dic['source'] = "TODO"  # self.source
            dic['source_cells'] = ""  # self.csv_rows

        # include statistics
        if self.opts.include_stats:
            dic['stats'] = self.stats()

        return dic

    def to_json(self):
        """Return the document data in :ref:`json` format

        .. code-block:: python

           ogtDoc = OGTDocument()
           ogtDoc.add_ags4_file("/my.ags")
           ogtDoc.opts.minify = False
           ogtDoc.opts.extended = False
           ags_str, err = ogtDoc.to_json()

        :rtype: (str or None, str)
        :return: A tuple with:

                - `None` if error else a `str` with :ref:`json` encoded data
                - An `error` string is error occured, else `None`
        """
        return utils.to_json(self.to_dict(), minify=self.opts.minify)

    def to_yaml(self):
        """Return the document data in :ref:`yaml` format

        :rtype: (str or None, str)
        :return: A tuple with:

                - `None` if error else a `str` with :ref:`yaml` encoded data
                - An `error` string is error occured, else `None`
        """
        return utils.to_yaml(self.to_dict())

    def get_points(self):
        """"""
        grpLoca = self.group_from_code("LOCA")
        if grpLoca is None:
            return
        lst = []

        # WSG84
        if grpLoca.has_heading("LOCAL_LAT") and grpLoca.has_heading("LOCAL_LON"):
            for rec in grpLoca.data:
                lat_s = rec.get("LOCA_LAT")
                lon_s = rec.get("LOCA_LON")
                # TODO add Point

        # BNG British National grid
        elif grpLoca.has_heading("LOCA_NATE") and grpLoca.has_heading("LOCA_NATN"):
            for rec in grpLoca.data:

                loca_id = rec.get("LOCA_ID")
                east = float(rec.get("LOCA_NATE"))
                north = float(rec.get("LOCA_NATN"))
                if east and north:
                    lat, lon = bng_to_latlon.OSGB36toWGS84(east, north)
                    lst.append(dict(lat=lat, lon=lon, east=east, north=north, loca_id=loca_id))
        return lst

    def to_geojson(self, minify=False):
        """geojson dict"""
        loca = self.group_from_code("LOCA")
        if loca is None:
            return None, "No `LOCA` Group"

        def make_feature(rec, lat, lon):
            props = dict(PointID=rec.get("LOCA_ID"), Type=rec.get("LOCA_TYPE"),
                         GroundLevel=rec.get("LOCA_GL"))
            return Feature(geometry=Point((lon, lat)), properties=props)

        features = []

        # WSG84
        if "LOCA_LAT" in loca.headings and "LOCA_LON" in loca.headings:
            for rec in loca.data:
                lat_s = rec.get("LOCA_LAT")
                lon_s = rec.get("LOCA_LON")
                if lat_s and lon_s:
                    features.append(make_feature(rec, lat_s, lon_s))

        # BNG British National grid
        elif "LOCA_NATE" in loca.headings and "LOCA_NATN" in loca.headings:
            for rec in loca.data:
                east = utils.to_int(rec.get("LOCA_NATE"))
                north = utils.to_int(rec.get("LOCA_NATN"))
                if east and north:
                    lat, lon = bng_to_latlon.OSGB36toWGS84(east, north)
                    features.append(make_feature(rec, lat, lon))

        if len(features) > 0:
            feature = FeatureCollection(features)
            return utils.to_json(feature, minify=minify)
        return None, None

    def write_excel(self, file_path):
        """Write out xlsx

        :param file_path: Full path of file to write
        :tyoe file_path: str
        :rtype: (str or None, str)
        :return: A tuple with:

                - `None` if error else a `str` with :ref:`json` encoded data
                - An `error` string is error occured, else `None`
        """
        wbook = Workbook()

        for gidx, grp_code in enumerate(self.groups_index):
            grpObj = self.groups_dict[grp_code]

            if gidx == 0:
                # By default an empty workbook has a first sheet
                sheet = wbook.active
                sheet.title = grp_code
            else:
                sheet = wbook.create_sheet(title=grp_code)

            for hidx, hd in enumerate(grpObj.headings_list()):

                # HEADINGS's
                sheet["A1"] = ags4.DESCRIPTOR.heading
                cell_ref = "%s%s" % (get_column_letter(hidx + 2), 1)
                sheet[cell_ref] = hd.head_code

                # TYPE
                sheet["A2"] = ags4.DESCRIPTOR.data_type
                cell_ref = "%s%s" % (get_column_letter(hidx + 2), 2)
                sheet[cell_ref] = hd.data_type

                # UNIT
                sheet["A3"] = ags4.DESCRIPTOR.unit
                cell_ref = "%s%s" % (get_column_letter(hidx + 2), 3)
                sheet[cell_ref] = hd.unit

                # DATA
                didx = 4
                for drow in grpObj.data_rows_dict():
                    sheet["A%s" % didx] = ags4.DESCRIPTOR.data
                    cell_ref = "%s%s" % (get_column_letter(hidx + 2), didx)
                    sheet[cell_ref] = drow[hd.head_code]

                    didx += 1
        try:
            wbook.save(file_path)
            return "Wrote `%s`" % file_path, None
        except Exception as error:
            return "Error: %s writing excel `%s`" % (error, file_path)

    def stats(self):
        """Statistics TODO"""
        dic = {}

        # Number of locations
        locaGrp = self.groups_dict.get("LOCA")
        if locaGrp is None:
            dic['locations'] = None
        else:
            recs = locaGrp.data_column_from_head_code("LOCA_ID")
            dic['locations'] = dict(count=len(recs), data=recs)

        # Data rows
        lst = []
        for group_key in sorted(self.groups_dict.keys()):
            grp = self.groups_dict.get(group_key)
            lst.append(dict(GROUP=group_key, count=len(grp.data)))
        dic['data'] = lst

        # Sample Types
        grp = self.groups_dict.get("SAMP")
        if not grp:
            dic['sample_types'] = None
        else:
            d = {}
            reccells = grp.data_column_from_head_code("SAMP_TYPE")
            recs = [cell.value for cell in reccells]

            for st in sorted(recs):
                if st not in d:
                    d[st] = 0
                d[st] += 1
            dic['sample_types'] = d

        # Site Geom
        d = {}
        # TODO X.Y.Z
        d['LOCA_LOCX'] = "todo"
        d['LOCA_LOCY'] = "todo"
        d['LOCA_LOCZ'] = "todo"

        # National Grid
        def calc_ng_stats(recs):
            # TODO - need to check type casting ?
            if recs is None:
                return None

            ds = {}
            ds['min'] = None if len(recs) == 0 else min(recs)
            ds['max'] = None if len(recs) == 0 else max(recs)
            ds['row_count'] = len(recs)
            ds['rows_with_data'] = 0
            ds['rows_without_data'] = 0
            for rec in recs:
                if not rec:
                    ds['rows_without_data'] += 1
                else:
                    ds['rows_with_data'] += 1
            return ds

        dic['site_geometry'] = None
        if locaGrp:
            recs = locaGrp.data_column_from_head_code("LOCA_NATE", as_cells=False)
            d['LOCA_NATE'] = calc_ng_stats(recs)

            recs = locaGrp.data_column_from_head_code("LOCA_NATN", as_cells=False)
            d['LOCA_NATN'] = calc_ng_stats(recs)

            recs = locaGrp.data_column_from_head_code("LOCA_GL", as_cells=False)
            d['LOCA_GL'] = calc_ng_stats(recs)

            dic['site_geometry'] = d

        # GEOL
        grp = self.groups_dict.get("GEOL")
        if not grp:
            dic['geol'] = None
        else:
            recs = grp.data_column("LOCA_ID")
            locs = dic['locations']['data']
            ll = []
            for l in locs:
                if l not in recs:
                    if l not in ll:
                        ll.append(l)
            dic['geol'] = dict(no_entries=ll if len(ll) > 0 else None)

        # SAMP
        grp = self.groups_dict.get("SAMP")
        if not grp:
            dic['samp'] = None
        else:
            recs = grp.data_column_from_head_code("LOCA_ID")
            locs = dic['locations']['data']
            ll = []
            for l in locs:
                if l not in recs:
                    if l not in ll:
                        ll.append(l)
            dic['samp'] = dict(no_entries=ll if len(ll) > 0 else None)

        # Unused Groups
        all_g = ags4.DD.groups_dict()
        dic['unused_groups'] = None
        ags_groups = all_g.keys()
        dic['unused_groups'] = sorted(list(set(ags_groups) - set(self.groups_dict.keys())))

        return dic

    def add_ags4_file(self, ags4_file_path):
        """Loads document from an :term:`ags4` formatted file

        :param ags4_file_path: absolute or relative path to file, will be at source_file_path
        :type ags4_file_path: str
        :rtype: str or None
        :return: A str if an error else None


        .. todo:: deal with utf-8 and windows.chars as ags is ascii only
                  another BIG issue is adding multiple files.. + source...s
        """
        # This function read's the file and call `add_ags4_string`

        try:
            with open(ags4_file_path, "r") as file:
                contents = file.read()
                self.add_ags4_string(contents, ags4_file_path)
                return None

        except IOError as error:
            return "%s - `%s`" % (str(error), ags4_file_path)

        # should never happen
        return "WTF in `add_ags4_file`"

    @property
    def source_file_path(self):
        """TODO wourkaround atmo"""
        return self.source_refs[0]

    def add_ags4_string(self, ags4_str, src_ref):
        """Load  document from an :term:`ags4` formatted string

        :param ags4_str: ags4 formatted string to load
        :type ags4_str: str
        :param src_ref: filename, eg excel, ags etc TODO
        :type src_ref: str
        :rtype: str or None
        :return: Any error  else `None`
        """

        self.source_refs.append(src_ref)
        self.source_strings[src_ref] = ags4_str

        # Note: the csv parser cant deal columns of different length,
        # Step 1:
        #  - split ags_string into lines
        #  - and parse `each line` into csv
        xcells = []
        for lidx, line in enumerate(ags4_str.split("\n")):

            # removing and trailing whitespace eg \r
            # were in nix land, so reassemble with CRLF when dumping to ags
            stripped = line.strip()

            if stripped == "":
                continue

            # decode csv line
            reader = csv.reader(StringIO(stripped))
            for row in reader:
                csv_row = row
                break

            # walk row and make OgtCell outta each bit of stuff
            row_cells = []
            for cidx, val in enumerate(csv_row):
                row_cells.append(OGTCell(src_ref=src_ref, lidx=lidx, cidx=cidx, raw=val))
            xcells.append(row_cells)
        row_cells = None

        # --------------------
        # next walk and clean

        def row_2_upper(rcells):
            for cell in rcells:
                ogt_validate.to_upper(cell)

        loop_group = None
        for lidx, row_cells in enumerate(xcells):
            lenny = len(row_cells)
            if lenny == 0:
                continue

            # validate its a descriptor
            dcell = row_cells[0]
            valid = True
            if not valid:
                continue
            else:
                ogt_validate.to_upper(dcell)
                descriptor = dcell.value

                if descriptor == ags4.DESCRIPTOR.group:
                    row_2_upper([row_cells[1]])
                    loop_group = self.add_get_group(row_cells[1], src_ref=src_ref)

                elif descriptor == ags4.DESCRIPTOR.heading:
                    row_2_upper(row_cells[1:])
                    head_list = loop_group.add_get_headings_row(row_cells[1:], src_ref=src_ref)

                elif descriptor == ags4.DESCRIPTOR.unit:
                    for cidx, lcell in enumerate(row_cells[1:]):
                        lcell.parentHeading = loop_group.headings_dict[head_list[cidx]]
                        loop_group.headings_dict[head_list[cidx]].set_unit(lcell)

                elif descriptor == ags4.DESCRIPTOR.data_type:
                    row_2_upper(row_cells[1:])
                    for cidx, lcell in enumerate(row_cells[1:]):
                        lcell.parentHeading = loop_group.headings_dict[head_list[cidx]]
                        loop_group.headings_dict[head_list[cidx]].set_data_type(lcell)

                elif descriptor == ags4.DESCRIPTOR.data:
                    datarow = {}
                    for cidx, lcell in enumerate(row_cells[1:]):
                        lcell.parentHeading = loop_group.headings_dict[head_list[cidx]]
                        datarow[head_list[cidx]] = lcell
                    loop_group.data.append(datarow)

                else:
                    print("LOST row", row_cells)
                    # loop_group.add_lost_row(rrow, lidx)

        self.set_sort_mode(self.SORT.recommended)
        self.generate_samp_ids()

        return None

    def generate_samp_ids(self):
        """"""
        # attempt to SAMP_ID
        # grp = self.group("SAMP")
        for gcode, grp in self.groups_dict.items():
            # if grp:
            for headObj in grp.headings_dict.values():
                if headObj.head_code == "SAMP_ID":
                    for ridx in range(0, grp.data_rows_count):
                        cells_dic = grp.data[ridx]
                        id_cell = cells_dic.get("SAMP_ID")

                        if id_cell is None:
                            # panix
                            continue
                        delim = "~"
                        if not id_cell.raw:
                            s = "@"
                            cell = cells_dic.get("LOCA_ID")
                            s += "?loca_id?" if cell is None else cell.value
                            s += delim

                            cell = cells_dic.get("SAMP_TYPE")
                            s += "?type?" if cell is None else cell.value
                            s += delim

                            cell = cells_dic.get("SAMP_REF")
                            if cell and cell.value:
                                s += "ref:" + cell.value
                            else:
                                cell = cells_dic.get("SAMP_TOP")
                                if cell:
                                    s += "top:" + cell.value
                                else:
                                    s += "ridx: %s" % ridx

                            id_cell.value = str(s)


def create_doc_from_ags4_file(ags_file_path):
    """Convenience function to create and load an OGTDocument from an ags file

    .. code-block:: python

        doc, err = ogt_doc.create_doc_from_ags4_file("/path/to/my.ags")
        if err:
            print err
        else:
            print doc.group("PROJ")
    """
    doc = OGTDocument()
    err = doc.add_ags4_file(ags_file_path)
    return doc, err


def create_doc_from_ags4_string(ags4_string, file_name=None):
    """Convenience function to create and load an OGTDocument from an ags string

    .. code-block:: python

        doc, err = ogt_doc.create_doc_from_ags4_string(ags4_string)
        if err:
            print err
        else:
            print doc.group("PROJ")
    """
    doc = OGTDocument()
    doc.add_ags4_string(ags4_string, file_name)
    return doc, None


def create_doc_from_json_file(json_file_path):
    """Creates a document from a :ref:`json` formatted file

    .. code-block:: python

        doc, err = ogt_doc.create_doc_from_json_file("/path/to/my.json")
        if err:
            print err

    :param json_file_path: absolute or relative path to file
    :type json_file_path: str
    :rtype: tuple
    :return: A `tuple` containing

        - An :class:`~ogt.ogt_doc.OGTDocument` object on success, else `None`
        - An `Error` message if error, otherwise `None`
    """
    data, err = utils.read_json_file(json_file_path)
    if err:
        return None, err

    groups = data.get('groups')
    if groups is None:
        return None, "Error: no `groups` key in json file"

    doc = OGTDocument()


    for group_code in groups.keys():

        group = groups[group_code]

        grp = OGTGroup(group_code)
        # TODO doc.append_group(grp)

        # add units + also headings
        for head_code in group['UNIT'].keys():
            valu = str(group['UNIT'][head_code])
            # TODO grp.units[head_code] = valu
            # TODO grp.headings[head_code] = valu

        # add TYPE
        for head_code in group['TYPE'].keys():
            valu = str(group['TYPE'][head_code])
            # TODO grp.data_types_dict[head_code] = valu

        # add data
        for rec in group['DATA']:
            dic = {}
            for head_code in rec.keys():

                dic[head_code] = str(rec[head_code])

            grp.data.append(dic)

    return doc, None


GROUPS_FIRST_CODES = ["PROJ", "TRAN", "LOCA", "HDPH", "GEOL", "SAMP"]
"""The groups to appear first, see :func:`groups_sort` """

GROUPS_LAST_CODES = ["ABBR", "DICT", "FILE", "TYPE", "UNIT"]
"""The groups to appear last, see :func:`groups_sort`"""

GROUPS_REQUIRED = ["PROJ", "TRAN", "TYPE", "UNIT", "ABBR"]
"""Mandatory groups"""

GROUP_HEADINGS_DEFAULT = {
    "PROJ": ["PROJ_ID", "PROJ_NAME"],
    "TRAN": ["TRAN_DATE", "TRAN_STAT"],
    "TYPE": ["TYPE_TYPE", "TYPE_DESC"],
    "UNIT": ["UNIT_UNIT", "UNIT_DESC"],
    "ABBR": ["ABBR_HDNG",  "ABBR_CODE", "ABBR_DESC", "ABBR_LIST"]
}
"""Default headings to add to :attr:`GROUPS_REQUIRED`"""


def groups_sort(unsorted_groups):
    """Returns a 'preferred' order for the group keys

    .. note::

        This is a bit hacky as the ags spec does not define
        any group order, however PROJ does sensibly tend to come first etc.

        So this return a list with

        - at the start - :attr:`GROUPS_FIRST_CODES`
        - at end - :attr:`GROUPS_LAST_CODES`
        - and everything else in the middle sorted alphabetically
        - One idea might be to push insitu stuff to front, schedule
          and lab stuff next, monitoring stuff to end to reflect project progression

    :return: group codes in preffered order
    :rtype: list[str]
    """

    # make a newlist/copy of the groups in doc
    grps = unsorted_groups[:]

    # the start groups
    ret_start = []
    for group_code in GROUPS_FIRST_CODES:
        if group_code in grps:
            # exists so add to start, and remove from list
            ret_start.append(group_code)
            grps.remove(group_code)

    # the end groups
    ret_end = []
    for group_code in GROUPS_LAST_CODES:
        if group_code in grps:
            # exists so add to end, and remove from list
            ret_end.append(group_code)
            grps.remove(group_code)

    return ret_start + sorted(grps) + ret_end



#!/usr/bin/env python3
# coding: utf-8
""""""
import sys
from argparse import ArgumentParser

from ogt import FORMATS
from ogt.ogt_doc import create_doc_from_json_file, create_doc_from_ags4_file
from ogt import ags4 #validate_ags4_file, update, report_to_string


"""CLI Main Parser"""
parser = ArgumentParser(description="Toolkit for geotechnical")

sub_parsers = parser.add_subparsers(help="commands", dest="command")


#=======================
## Convert Parser
p_convert = sub_parsers.add_parser("convert", help="Convert files")

p_convert.add_argument("-b", "--beside", dest="beside", action="store_true",
                       help="save beside original file and appends `.ext`")

p_convert.add_argument("-e", "--extended", dest="extended", action="store_true",
                       help="Outputs header/group data, eg descriptions, picklists, etcs")

p_convert.add_argument("-f", "--format", dest="format", default="json",
                       help="output format and ext", choices=FORMATS)

p_convert.add_argument("-m", "--minify", dest="minify", action="store_true",
                       help="minify the output by removing whitespace")

p_convert.add_argument("-o", "--overwrite", dest="overwrite", action="store_true",
                       help="overwrite output file if it exists")

p_convert.add_argument("--stats", dest="inc_stats", action="store_true",
                       help="include stats in output (only json/yaml)")
p_convert.add_argument("--source", dest="inc_source", action="store_true",
                       help="include source text and cells")

p_convert.add_argument("-w", "--write_file", dest="write_file", help="output file to write")

p_convert.add_argument("-z", "--zip", dest="zip", action="store_true",
                       help="output as a zip file containing original and converted file")

p_convert.add_argument("source_file", type=str, help="AGS4 file to convert")

#=======================
## Validator parser
p_validate = sub_parsers.add_parser("validate", help="Validate files")
p_validate.add_argument("-p", action="store_false", dest="printable",
                        help="Human print output")
p_validate.add_argument("-r", "--rules", type=int, dest="rules", nargs='+', default=[],
                        help="Rules to check")
p_validate.add_argument("-t", "--tests", action="store_true", dest="run_tests",
                        help="Run Tests")
p_validate.add_argument("source_file", type=str, help="AGS4 file to validate")



## Add some standard items to all commands
for pra in [p_convert, p_validate]:
    pra.add_argument("--debug", dest="debug", action="store_false", help="debug")

if __name__ == "__main__":
    args = parser.parse_args()



    if args.command == "validate" and args.source_file.endswith(".ags"):
        report, error = ags4.validate_ags4_file(args.source_file, args.rules)

        if error is not None:
            print(error)
            sys.exit(0)

        if args.printable:
            print(ags4.report_to_string(report))
        else:
            print(report)

    elif args.command == "convert":
        if args.source_file.endswith(".json"):
            doc, error = create_doc_from_json_file(args.source_file)
        elif args.source_file.endswith(".ags"):
            doc, error = create_doc_from_ags4_file(args.source_file)

        if error is not None:
            print(error)
            sys.exit(0)

        doc.opts.minify = args.minify
        doc.opts.extended = args.extended
        doc.opts.include_stats = args.inc_stats
        doc.opts.include_source = args.inc_source
        message, error = doc.write(ext=args.format, beside=args.beside,
                                   file_path=args.write_file,
                                    to_zip=args.zip, overwrite=args.overwrite)
        if error:
            print(error)
            sys.exit(0)

        print(message)
        sys.exit(0)


    else:
        print("Nothings to do !\n")
        parser.print_help()
        sys.exit(0)



# -*- coding: utf-8 -*-

from Qt import Qt, QtCore, QtWidgets, QtGui, pyqtSignal

from ogt import ogt_cell

import app_globals as G
import ogtgui_widgets
import xwidgets

class SourceGridModel(QtCore.QAbstractTableModel):

    sigChanged = pyqtSignal()

    def __init__(self, parent=None, ogtDoc=None):
        QtCore.QAbstractTableModel.__init__(self, parent)

        self.ogtDoc = ogtDoc

        self.xcells = None
        self.max_cols = None

    def rowCount(self, parent=None, *args):
        """Returns the number of rows of the model"""
        if self.xcells == None:
            return 0
        return len(self.xcells)

    def columnCount(self, parent=None, *args):
        """Returns the number of columns of the model"""
        if self.max_cols != None:
            return self.max_cols
        return 0

    def DEADflags(self, midx):

        #print(midx)
        cell = self.xcells[midx.row()].get(midx.column())
        #head = self.ogtGroup.heading_from_index(idx.column())
        #dt = head.data_type
        if cell and isinstance(cell, ogt_cell.OGTCell) and cell.parentHeading:
            if cell.parentHeading.data_type == "PA":
            #if dt == "PA":  # or dt == "ID"
                flags = Qt.ItemIsSelectable | Qt.ItemIsEnabled

            else:
                flags = Qt.ItemIsSelectable | Qt.ItemIsEnabled | Qt.ItemIsEditable
            # if index.column() != 0:
            #    flags |= Qt.ItemIsEditable
            return flags
        flags =  Qt.ItemIsEnabled
        return flags

    # def setData(self, index, value, role=None):
    #     """Updates DATA when modified in the view"""
    #     if role == Qt.EditRole:
    #         cell = self.ogtGroup.data_cell(index.row(), index.column())
    #         cell.value = value.strip()
    #         self.ogtGroup.emit()
    #         #success = self.ogtGroup.set_data_cell_value(index.row(), index.column(), value)
    #         #self.layoutChanged.emit()
    #         #self.modelReset.emit()
    #         #self.sigChanged.emit()
    #         return True
    #         #self.dataChanged.emit(index, index)
    #         #return True
    #     return False

    def data(self, index, role=Qt.DisplayRole):
        """Returns the data at the given index"""

        row = index.row()
        col = index.column()

        cell_val = self.xcells[row].get(col)

        if role == Qt.DisplayRole:
            if isinstance(cell_val, ogt_cell.OGTCell):
                return cell_val.value
            return cell_val if cell_val else ""

        if role == Qt.BackgroundColorRole:
            if cell_val:
                if isinstance(cell_val, ogt_cell.OGTCell):
                    # all the data is cells
                    return QtGui.QColor(cell_val.background())
                # the descriptions are string
                if cell_val == "DATA":
                    return QtGui.QColor("#DCF1FA")
                elif cell_val == "GROUP":
                    return QtGui.QColor("#78DAFF")
                else:
                    return QtGui.QColor("#ABDBFE")
            return QtGui.QColor("#dddddd")

        return None

    def headerData(self, idx, orientation, role=Qt.DisplayRole):

        if orientation == Qt.Horizontal and role == Qt.DisplayRole:
            return str(idx + 1)

        if orientation == Qt.Vertical and role == Qt.DisplayRole:
            return str(idx + 1)

        return None

    def do_update(self):

        self.xcells, self.max_cols = self.ogtDoc.to_grid()
        self.modelReset.emit()


class SourceGridTableWidget( QtWidgets.QWidget ):

    #sigGoto = pyqtSignal(str)
    sigSelect = pyqtSignal(object)

    def __init__( self, parent=None, ogtDoc=None):
        QtWidgets.QWidget.__init__( self, parent )


        self.ogtDoc = ogtDoc

        self.mainLayout = xwidgets.vlayout()
        self.setLayout(self.mainLayout)


        ## Sources Table --------------
        self.tableGrid = QtWidgets.QTableView()
        self.mainLayout.addWidget(self.tableGrid, 0)


        self.modelGrid = SourceGridModel(ogtDoc=self.ogtDoc)
        self.tableGrid.setModel(self.modelGrid)

        f = self.tableGrid.font()
        f.setFamily("monospace")
        self.tableGrid.setFont(f)
        verticalHeader = self.tableGrid.verticalHeader()
        #TODO verticalHeader.setResizeMode(QtWidgets.QHeaderView.Fixed)
        verticalHeader.setDefaultSectionSize(20)


    def do_update(self):
        self.modelGrid.do_update()



class OGTSourceViewWidget( QtWidgets.QWidget ):
    """The SourceViewWidget info which in row 0 """

    def __init__( self, parent=None, ogtDoc=None):
        QtWidgets.QWidget.__init__( self, parent )


        self.ogtDoc = ogtDoc
        self.setObjectName("OGTSourceViewWidget")

        self.mainLayout = QtWidgets.QVBoxLayout()
        self.mainLayout.setSpacing(0)
        self.mainLayout.setContentsMargins(0,0,0,0)
        self.setLayout(self.mainLayout)

        self.headerWidget = ogtgui_widgets.HeaderBarWidget(title="Source Output", bg="#B2C3C3")
        self.mainLayout.addWidget(self.headerWidget, 0)

        self.mainLayout.addStretch(5)

        self.tabWidget = QtWidgets.QTabWidget()
        self.mainLayout.addWidget(self.tabWidget, 200)

        # Source View
        self.sourceView = xwidgets.XPlainTextEdit()
        self.tabWidget.addTab(self.sourceView, "Raw Text")

        # Grid view
        self.tableGrid = SourceGridTableWidget(ogtDoc=self.ogtDoc)
        self.tabWidget.addTab(self.tableGrid, "Grid View")


        self.tabWidget.setCurrentIndex(1)


    def do_update(self):


        txt, err = self.ogtDoc.to_ags4()
        self.sourceView.setPlainText(txt)

        self.tableGrid.do_update()


    def on_splitter_moved(self, i, pos):
        G.settings.save_splitter(self.splitter)

    def load_document(self, doco):

        self.sourceView.setText(doco.source)

        show_warn, show_err = self.errorsWidget.get_error_filters()
        self.tableWidget.setRowCount(len(doco.csv_rows))

        for ridx, row in enumerate(doco.csv_rows):

            # each csv_row is not the same, so extend here
            if self.tableWidget.columnCount() < len(row):
                self.tableWidget.setColumnCount(len(row))


            #errs = doco.error_cells.get(ridx)

            bg = None
            for cidx, cell in enumerate(row):

                #if cidx == 0:
                #    bg = bg_color(cell)
                item = xwidgets.XTableWidgetItem()
                item.setText( cell )
                item.set_bg("#EAFFE0")
                self.tableWidget.setItem(ridx, cidx, item)


                errs = doco.get_errors(lidx=ridx, cidx=cidx)
                if errs != None:

                    for er in errs:

                        if er.type:
                            item.setData(FILTER_ROLE.err, "1")
                        else:
                            item.setData(FILTER_ROLE.warn, "1")
                        """    
                        if er.cidx == cidx:
                            item.set_bg(er.bg)
                        """
                    self.set_item_bg(item, show_warn, show_err)
                ## color the row
                #item.setBackgroundColor(QtWidgets.QColor(bg))
            self.tableWidget.setRowHeight(ridx, 20)


        self.errorsWidget.load_document(doco)

    def set_item_bg(self, item, show_warn, show_err):
        has_warn = item.data(FILTER_ROLE.warn).toBool()
        has_err = item.data(FILTER_ROLE.err).toBool()
        item.set_bg("white")
        if show_warn and has_warn:
            item.set_bg(ERR_COLORS.warn_bg)
        if show_err and has_err:
            item.set_bg(ERR_COLORS.err_bg)


    def update_colours(self, show_warn, show_err):
        self.tableWidget.setUpdatesEnabled(False)
        for ridx in range(0, self.tableWidget.rowCount()):
            for cidx in range(0, self.tableWidget.columnCount()):
                item = self.tableWidget.item(ridx, cidx)
                if item:
                    self.set_item_bg(item, show_warn, show_err)
        self.tableWidget.setUpdatesEnabled(True)

    def select_cell(self, lidx, cidx):
        self.tabWidget.setCurrentIndex(self.tabWidget.indexOf(self.tableWidget))
        self.tableWidget.setCurrentCell(lidx, cidx)
        item = self.tableWidget.currentItem()
        self.tableWidget.scrollToItem(item, QtWidgets.QAbstractItemView.PositionAtCenter)


    def on_select_changed(self):
        item = self.tableWidget.currentItem()
        if item == None:
            self.errorsWidget.select_items(None, None)
            return

        self.errorsWidget.select_items(item.row(), item.column())


    def on_row_clicked(self, ridx):
        pass




# -*- coding: utf-8 -*-


from Qt import QtGui, QtWidgets, QtCore, Qt, pyqtSignal



from img import Ico
import xwidgets

class OgtHomePortal( QtWidgets.QWidget ):

    #sigFileSelected = pyqtSignal(object)

    def __init__( self, parent):
        QtWidgets.QWidget.__init__( self, parent )

        self.debug = False

        self.setMinimumWidth(300)

        self.mainLayout = QtWidgets.QVBoxLayout()
        self.mainLayout.setSpacing(0)
        self.mainLayout.setContentsMargins(0,0,0,0)
        self.setLayout(self.mainLayout)


        self.projectsWiget = OgtProjectsWidget()
        self.mainLayout.addWidget(self.projectsWiget)

class C_EG:
    """Columns for examples"""
    file_name = 0

class OgtProjectsWidget( QtWidgets.QWidget ):

    sigFileSelected = pyqtSignal(object)

    def __init__( self, parent):
        QtWidgets.QWidget.__init__( self, parent )

        self.debug = False

        self.setMinimumWidth(300)

        self.mainLayout = QtWidgets.QVBoxLayout()
        self.mainLayout.setSpacing(0)
        self.mainLayout.setContentsMargins(0,0,0,0)
        self.setLayout(self.mainLayout)


        #=============================
        ## Set up tree
        self.tree = QtWidgets.QTreeWidget()
        self.mainLayout.addWidget(self.tree, 30)

        self.tree.setRootIsDecorated(False)
        self.tree.header().setStretchLastSection(True)
        self.tree.header().hide()

        hi = self.tree.headerItem()
        hi.setText(C_EG.file_name, "Example")


        self.tree.itemClicked.connect(self.on_tree_item_clicked)

        self.load_projects()



    def load_projects(self, sub_dir=None):
        return
        files_list, err = ags4.examples_list()
        if err:
            pass #TODO
        self.tree.clear()

        for fd in files_list:
            file_name = fd["file_name"]
            item = QtWidgets.QTreeWidgetItem()
            item.setText(C_EG.file_name, file_name)
            item.setIcon(C_EG.file_name, Ico.icon(Ico.Ags4))
            f = item.font(C_EG.file_name)
            f.setBold(True)
            item.setFont(C_EG.file_name, f)
            self.tree.addTopLevelItem(item)



    def on_tree_item_clicked(self, item, col):

        file_name = str(item.text(C_EG.file_name))
        self.sigFileSelected.emit(file_name)




import math

from Qt import Qt, QtCore, QtGui, QtWidgets, QtChart, pyqtSignal

from ogt import utils
import ogtgui_widgets
import xwidgets
from img import Ico

SEL_COLOR = "#F5FFA7"

class OGTSamplesModel(QtCore.QAbstractItemModel):


    def __init__(self, parent=None, ogtDoc=None):
        QtCore.QAbstractItemModel.__init__(self)

        self.ogtDoc = ogtDoc
        self.loca_id = None


    def index(self, row, col, parent=None):
        return self.createIndex(row, col)

    def parent(self, pidx=None):
        return QtCore.QModelIndex()

    def row_index_from_samp_id(self, samp_id):
        grp = self.ogtDoc.group_from_code("SAMP")
        #cidx = grp.index_of_heading("SAMP_ID")
        for ridx, row in enumerate(grp.data):
            cell = row.get("SAMP_ID")
            if cell.value == samp_id:
                return ridx
        return None

    def get_index_from_cell(self, scell):
        grp = self.ogtDoc.group_from_code("SAMP")
        for ridx in range(0, grp.data_rows_count):
            for cidx in range(0, grp.headings_count):
                cell = grp.data_cell(ridx, cidx)
                if cell == scell:
                    return self.index(ridx, cidx)
        return None


    def columnCount(self, pidx=None):
        grp = self.ogtDoc.group_from_code("SAMP")
        if grp:
            return grp.headings_count
        return 0

    def rowCount(self, pidx=None):
        grp = self.ogtDoc.group_from_code("SAMP")
        if grp:
            return grp.data_rows_count
        return 0

    def data_row(self, row):
        return self.ogtDoc.group_from_code("SAMP").data[row]

    def data(self, midx, role=Qt.DisplayRole):
        """Returns the data at the given index"""
        row = midx.row()
        col = midx.column()

        if role == Qt.DisplayRole or role == Qt.EditRole:
            cell = self.ogtDoc.group_from_code("SAMP").data_cell(row, col)
            return cell.value




        #if role == Qt.DecorationRole:
        #    if self.ogtDoc.group_from_code("SAMPE").heading_from_index(col).head_code == "SAMP_ID":
        #        return Ico.icon(Ico.Samples)

        #if role == Qt.FontRole:
        #    if col == self.C.group_code:
        #        f = QtWidgets.QFont()
        #        f.setBold(True)
        #        f.setFamily("monospace")
        #        return f

        #if role == Qt.TextAlignmentRole:
        #    return Qt.AlignRight if col == 0 else Qt.AlignLeft

        if role == Qt.BackgroundRole:
            bg = None
            if self.loca_id:
                xloca = self.ogtDoc.group_from_code("SAMP").data[row].get("LOCA_ID")
                if xloca and xloca.value == self.loca_id:
                    bg = SEL_COLOR

            return QtGui.QColor(bg) if bg != None else None


        return None


    def headerData(self, idx, orient, role=None):
        if role == Qt.DisplayRole and orient == Qt.Horizontal:
            #heads = self.ogtDoc.group_from_code("SAMP").heading_from_index(idx)
            return self.ogtDoc.group_from_code("SAMP").heading_from_index(idx).head_code

        if role == Qt.TextAlignmentRole and orient == Qt.Horizontal:
            return Qt.AlignRight if idx == 0 else Qt.AlignLeft




class OGTTestPointsModel(QtCore.QAbstractItemModel):


    def __init__(self, parent=None, ogtDoc=None):
        QtCore.QAbstractItemModel.__init__(self)

        self.ogtDoc = ogtDoc
        self.loca_id = None


    def index(self, row, col, parent=None):
        return self.createIndex(row, col)

    def parent(self, pidx=None):
        return QtCore.QModelIndex()


    def columnCount(self, foo):
        grp = self.ogtDoc.group_from_code("LOCA")
        if grp:
            return grp.headings_count
        return 0

    def index_from_locaid(self, loca_id):
        grp = self.ogtDoc.group_from_code("LOCA")
        if grp:
            for ridx in range(0, grp.data_rows_count):
                cell = grp.data[ridx].get("LOCA_ID")
                if cell.value == loca_id:
                    return self.index(ridx, grp.index_of_heading("LOCA_ID"))
    def rowCount(self, midx):
        grp = self.ogtDoc.group_from_code("LOCA")
        if grp:
            return grp.data_rows_count
        return 0

    def data(self, midx, role=Qt.DisplayRole):
        """Returns the data at the given index"""
        row = midx.row()
        col = midx.column()

        if role == Qt.DisplayRole:
            #grp = self.ogtDoc.group_from_code("LOCA")
            #hd = grp.heading_from_index(col)

            cell = self.ogtDoc.group_from_code("LOCA").data_cell(row, col)

            return cell.value

        if role == Qt.DecorationRole:
            if self.ogtDoc.group_from_code("LOCA").heading_from_index(col).head_code == "LOCA_ID":
                return Ico.icon(Ico.TestPoint)

        if False and  role == Qt.FontRole:
            if self.ogtDoc.group_from_code("LOCA").heading_from_index(col).head_code == "LOCA_ID":
                f = QtWidgets.QFont()
                f.setBold(True)
                f.setFamily("monospace")
                return f

        if  role == Qt.TextAlignmentRole:
            return Qt.AlignLeft

        if role == Qt.BackgroundColorRole:
            grp = self.ogtDoc.group_from_code("LOCA")
            if self.loca_id:
                cell = grp.data_cell(row, grp.index_of_heading("LOCA_ID"))
                if self.loca_id == cell.value:
                    return QtGui.QColor(SEL_COLOR)


            return None


        return QtCore.QVariant()


    def headerData(self, idx, orient, role=None):
        if role == Qt.DisplayRole and orient == Qt.Horizontal:
            #heads = self.ogtDoc.group_from_code("SAMP").heading_from_index(idx)
            grp = self.ogtDoc.group_from_code("SAMP")
            if grp == None:
                return ""
            return grp.heading_from_index(idx).head_code


        return None



class OGTSamplesWidget( QtWidgets.QWidget ):


    #sigGoto = pyqtSignal(str)

    def __init__( self, parent=None, ogtDoc=None):
        QtWidgets.QWidget.__init__( self, parent )

        self.debug = False
        self.ogtDoc = ogtDoc

        self.containerLayout = xwidgets.vlayout()
        self.setLayout(self.containerLayout)

        self.headerWidget = ogtgui_widgets.HeaderBarWidget(title="Samples", bg="#F7E194")
        self.containerLayout.addWidget(self.headerWidget, 0)

        self.mainLayout = xwidgets.hlayout()
        self.containerLayout.addLayout(self.mainLayout, 100)

        self.splitter = QtWidgets.QSplitter()
        self.mainLayout.addWidget(self.splitter)

        ## ========= Test Points
        self.treeLocations = QtWidgets.QTreeView()
        self.splitter.addWidget(self.treeLocations)

        self.modelTestPoints = OGTTestPointsModel(ogtDoc=self.ogtDoc)
        self.treeLocations.setModel(self.modelTestPoints)

        self.treeLocations.setRootIsDecorated(False)
        self.treeLocations.setExpandsOnDoubleClick(False)
        self.treeLocations.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
        self.treeLocations.setSelectionMode(QtWidgets.QAbstractItemView.SingleSelection)


        self.treeLocations.selectionModel().selectionChanged.connect(self.on_tree_locs_selection)

        ## ========= Samples
        self.treeSamples = QtWidgets.QTreeView()
        self.splitter.addWidget(self.treeSamples)

        self.modelSamples = OGTSamplesModel(ogtDoc=self.ogtDoc)
        self.treeSamples.setModel(self.modelSamples)

        self.treeSamples.setRootIsDecorated(False)
        self.treeSamples.setExpandsOnDoubleClick(False)
        self.treeSamples.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
        self.treeSamples.setSelectionMode(QtWidgets.QAbstractItemView.SingleSelection)

        self.treeSamples.selectionModel().selectionChanged.connect(self.on_tree_samples_selection_changed)


        chartContainerWidget = QtWidgets.QWidget()
        self.splitter.addWidget(chartContainerWidget)
        chartLayout = xwidgets.vlayout()
        chartContainerWidget.setLayout(chartLayout)

        lbl = xwidgets.XLabel(text="Chart below is a placeholder and needs stacked bar or alike",
                              style="background-color: #FFFCBF; padding: 5px")
        chartLayout.addWidget(lbl)

        self.chart = QtChart.QChartView()
        chartLayout.addWidget(self.chart)

        self.splitter.setStretchFactor(0, 0)
        self.splitter.setStretchFactor(1, 0)
        self.splitter.setStretchFactor(3, 6)


    def on_tree_samples_selection_changed(self, sel, dsel):

        self.treeLocations.blockSignals(True)
        self.treeLocations.clearSelection()
        self.treeLocations.blockSignals(False)

        indexes = self.treeSamples.selectionModel().selectedIndexes()
        loca_id = None



        self.modelTestPoints.loca_id = loca_id
        self.modelTestPoints.layoutChanged.emit()

        self.modelSamples.loca_id = loca_id
        self.modelSamples.layoutChanged.emit()



    def on_tree_locs_selection(self, sel, dsel):

        loca_id = None
        indexes = self.treeLocations.selectionModel().selectedIndexes()
        if len(indexes):
            loca_id = self.modelTestPoints.data(indexes[0], Qt.DisplayRole)



        self.modelSamples.loca_id = loca_id
        self.modelSamples.layoutChanged.emit()

        self.treeSamples.blockSignals(True)
        self.treeSamples.clearSelection()
        self.treeSamples.blockSignals(False)


    def do_update(self):
        self.modelTestPoints.layoutChanged.emit()
        self.modelSamples.layoutChanged.emit()

        self.series = {}
        max_d = 0
        locas = {}
        sampGrp = self.ogtDoc.group_from_code("SAMP")
        if sampGrp == None:
            return
        for rec in sampGrp.data:
            locaCell = rec.get("LOCA_ID")
            if locaCell:
                loca_id = locaCell.value
                if not loca_id in locas:
                    locas[loca_id] = []
                    self.series[loca_id] = []
                locas[loca_id].append(rec)


                dd = None
                st = rec.get("SAMP_TOP")
                if st:
                    dd = utils.to_float(st.value)
                    self.series[loca_id].append(dd)
                # rec['__DEPTH__'] = dd      # @@@DEPTH
                if dd != None and dd > max_d:
                    max_d = dd

        self.locas = sorted(locas.keys())

        self.chart.chart().removeAllSeries()


        axisX = QtChart.QBarCategoryAxis()
        axisX.append(self.locas)
        self.chart.chart().addAxis(axisX, Qt.AlignTop)

        axisY = QtChart.QValueAxis()
        self.chart.chart().addAxis(axisY, Qt.AlignLeft)


        for loca in self.locas:

            series = QtChart.QStackedBarSeries()
            series.setName(loca)

            barSet = QtChart.QBarSet(loca)
            for dpt in self.series[loca]:

                if dpt is not None:
                    barSet.append(dpt)

            series.append(barSet)

            self.chart.chart().addSeries(series)
            series.attachAxis(axisX)
            series.attachAxis(axisY)

        self.chart.setMinimumWidth(600)
        # self.chart.chart().createDefaultAxes()





#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import sys
import argparse
from PyQt5 import QtGui, QtWidgets

import ogtgui.main_window


parser = argparse.ArgumentParser(description="Launch desktop application")
parser.add_argument("--dev", help="Developer mode", action="store_true")

parser.add_argument("ags_file", nargs='?', type=str, help="AGS formatted file to load at startup")

if __name__ == '__main__':


    args = parser.parse_args()

    app = QtWidgets.QApplication( sys.argv )

    if args.dev == False:
        splashScreen = ogtgui.main_window.MainWindow.show_splash()
    app.processEvents()

    ## Main Window
    window = ogtgui.main_window.MainWindow(args)

    if args.dev == False:
        splashScreen.finish( window )
    window.show()

    sys.exit( app.exec_() )

"""
?? this autodoc not work ;-(
.. _ogt-desktop.py:

.. autoprogram:: ogt-desktop:parser
   :prog: ogt-desktop.py

.. seealso::

    - :mod:`ogtgui`

"""



# -*- coding: utf-8 -*-
"""
@author: Peter Morgan <pete@daffodil.uk.com>
"""
import os
from Qt import Qt, QtCore, QtGui, QtWidgets, pyqtSignal

from img import Ico

DEFAULT_SPACING = 0
DEFAULT_MARGIN = 0
DEFAULT_BUTTON_WIDTH = 80

#=====================================================
# Layouts

def hlayout(spacing=DEFAULT_SPACING, margin=DEFAULT_MARGIN):
    """Convenience function to create a QHBoxLayout"""
    lay = QtWidgets.QHBoxLayout()
    if isinstance(margin, bool):
        margin = DEFAULT_SPACING
    if isinstance(spacing, bool):
        spacing = DEFAULT_SPACING
    lay.setContentsMargins(margin, margin, margin, margin)
    lay.setSpacing(spacing)
    return lay

def vlayout(spacing=DEFAULT_SPACING, margin=DEFAULT_MARGIN):
    """Convenience function to create a QVBoxLayout"""
    lay = QtWidgets.QVBoxLayout()
    if isinstance(margin, bool):
        margin = DEFAULT_SPACING
    if isinstance(spacing, bool):
        spacing = DEFAULT_SPACING
    lay.setContentsMargins(margin, margin, margin, margin)
    lay.setSpacing(spacing)
    return lay


def file_exists_check(parent, file_path):
    if not os.path.exists(file_path):
        mess = QtWidgets.QMessageBox
        mess.critical(parent, "File not exist", file_path)
        return False
    return True

class XLabel(QtWidgets.QLabel):

    sigClicked = pyqtSignal()

    def __init__( self, parent=None, style=None,  align=None,
                 text=None,  tooltip=None, bold=False, width=None):
        QtWidgets.QLabel.__init__( self, parent )

        self.set_bold(bold)

        if text:
            self.setText(text)

        if tooltip:
            self.setToolTip(tooltip)

        if align != None:
            self.setAlignment(align)

        if style:
            self.setStyleSheet(style)

        if width != None:
            self.setFixedWidth(width)

    def set_bold(self, state):
        f = self.font()
        f.setBold(state)
        self.setFont(f)

    def mousePressEvent(self, ev):
        self.sigClicked.emit()
        QtWidgets.QLabel.mousePressEvent(self, ev)


class XToolButton(QtWidgets.QToolButton):
    def __init__( self, parent=None, both=True,  ico=None,
                  popup=False, autoRaise=True, menu=False, disabled=False,
                text=None,  tooltip=None, bold=False,
                  callback=None, toggledCallback=None,
                  checkable=None, checked=None,
                width=None):
        QtWidgets.QToolButton.__init__( self, parent )


        self.setAutoRaise(autoRaise)


        self.setDisabled(disabled)

        if width:
            self.setFixedWidth(width)

        if both:
            self.setToolButtonStyle( Qt.ToolButtonTextBesideIcon)
        else:
            self.setToolButtonStyle( Qt.ToolButtonIconOnly)

        if checkable != None:
            self.setCheckable(checkable)
        if checked != None:
            self.setChecked(checked)

        if callback:
            self.clicked.connect(callback)
        if toggledCallback:
            self.toggled.connect(toggledCallback)

        if tooltip:
            self.setToolTip(tooltip)

        if text:
            self.setText(text)

        if bold:
            self.setBold(True)

        if ico:
            self.setIco(ico)

        if popup:
            self.setPopupMode(QtWidgets.QToolButton.InstantPopup)

        if menu:
            self.setMenu(QtWidgets.QMenu())

    def setIco(self, ico):
        self.setIcon(Ico.icon(ico))

    def setBold(self, state=True):

        f = self.font()
        f.setBold(True)
        self.setFont(f)

    def set_bg(self, color):
        self.setStyleSheet("background-color: %s" % color)



class XPushButton(QtWidgets.QPushButton):
    def __init__( self, parent=None,  ico=None,
                 disabled=False,
                text=None,  tooltip=None, bold=False,
                  callback=None,       width=None):
        QtWidgets.QPushButton.__init__( self, parent )


        self.setDisabled(disabled)

        if width:
            self.setFixedWidth(width)



        if callback:
            self.clicked.connect(callback)

        if tooltip:
            self.setToolTip(tooltip)

        if text:
            self.setText(text)

        if bold:
            self.setBold(True)

        if ico:
            self.set_ico(ico)

    def set_ico(self, ico):
        self.setIcon(Ico.icon(ico))

    def setBold(self, state=True):

        f = self.font()
        f.setBold(True)
        self.setFont(f)

    def set_bg(self, color):
        self.setStyleSheet("background-color: %s" % color)

class XTableWidgetItem(QtWidgets.QTableWidgetItem):
    """Extended QTableWidgetItem with convenience functions"""
    def __init__(self):
        QtWidgets.QTableWidgetItem.__init__(self)



    def set(self, text=None, bold=False, bg=None, fg=None, align=None, check=None):

        if text:
            self.setText(text)

        if bold:
            self.set_bold(True)

        if bg:
            self.set_bg(bg)
        if fg:
            self.set_bg(fg)

        if align:
            self.setTextAlignment(align)

        if check != None:
            self.setCheckState(check)


    def set_bold(self, state):
        f = self.font()
        f.setBold(state)
        self.setFont(f)

    def set_bg(self, bg_color):
        colo = QtWidgets.QColor()
        colo.setNamedColor(bg_color)
        self.setBackgroundColor(colo)

    def set_fg(self, bg_color):
        colo = QtWidgets.QColor()
        colo.setNamedColor(bg_color)
        self.setForeground(colo)


class XTreeWidgetItem(QtWidgets.QTreeWidgetItem):
    """Extended QTableWidgetItem with convenience functions"""
    def __init__(self):
        QtWidgets.QTreeWidgetItem.__init__(self)


    def set(self, cidx, text=None, bold=False, bg=None, fg=None, align=None, check=None, ico=None):

        if text:
            self.setText(cidx, str(text))

        if bold:
            self.set_bold(cidx, True)

        if bg:
            self.set_bg(cidx, bg)
        if fg:
            self.set_fg(cidx, fg)

        if align:
            self.setTextAlignment(cidx, align)

        if check != None:
            self.setCheckState(cidx, check)

        if ico:
            self.set_ico(cidx, ico)


    def set_bold(self, cidx, state):
        f = self.font(cidx)
        f.setBold(state)
        self.setFont(cidx, f)

    def set_bg(self,cidx,  bg_color):
        colo = QtWidgets.QColor()
        colo.setNamedColor(bg_color)
        self.setBackgroundColor(cidx, colo)

    def set_fg(self, cidx,  bg_color):
        colo = QtWidgets.QColor()
        colo.setNamedColor(bg_color)
        self.setForeground(cidx, colo)

    def set_ico(self, cidx, ico):
        self.setIcon(cidx, Ico.icon(ico))

    def i(self, cidx):
        try:
            return int(str(self.text(cidx)))
        except:
            return None

class GroupHBox(QtWidgets.QGroupBox):


    def __init__(self, parent=None):
        QtWidgets.QGroupBox.__init__(self, parent)


        self.layout = QtWidgets.QHBoxLayout()
        self.setLayout(self.layout)

    def setContentsMargins(self, a,b,c,d):
        self.layout.setContentsMargins(a,b,c,d)

    def addWidget(self, widget, stretch=0):
        self.layout.addWidget(widget, stretch)

    def addLayout(self, widget, stretch=0):
        self.layout.addLayout(widget, stretch)

    def addStretch(self, stretch):
        self.layout.addStretch(stretch)



class GroupVBox(QtWidgets.QGroupBox):

    def __init__(self, parent=None, bold=False):
        QtWidgets.QGroupBox.__init__(self, parent)

        self.layout = QtWidgets.QVBoxLayout()
        self.layout.setSpacing(0)
        self.setLayout(self.layout)

    def setContentsMargins(self, a,b,c,d):
        self.layout.setContentsMargins(a,b,c,d)

    def addLabel(self, txt):
        lbl = QtWidgets.QLabel()
        lbl.setText(txt)
        lbl.setStyleSheet("font-family: monospace; font-size: 8pt; color: #666666; background-color: #efefef; padding: 3px;")
        self.layout.addWidget(lbl)


    def addWidget(self, widget, stretch=0):
        self.layout.addWidget(widget, stretch)

    def addLayout(self, widget, stretch=0):
        self.layout.addLayout(widget, stretch)

    def addSpacing(self, s):
        self.layout.addSpacing( s)

    def addStretch(self, stretch):
        self.layout.addStretch(stretch)


    def setSpacing(self, x):
        self.layout.setSpacing(x)

class GroupGridBox(QtWidgets.QGroupBox):


    def __init__(self, parent=None):
        QtWidgets.QGroupBox.__init__(self, parent)


        self.grid = QtWidgets.QGridLayout()
        self.setLayout(self.grid)



class XStandardItem( QtGui.QStandardItem ):

    def __init__( self):
        QtGui.QStandardItem.__init__( self )
        #super(QtWidgets.QStandardItem, self).__init__()

        self.setEditable(False)

    def set_bg(self, color):
        self.setBackground(  G.colors.to_object( color ) )

    def set_fg(self, color):
        self.setForeground(  G.colors.to_object( color ) )

    def set_bold( self, bold):
        font = self.font()
        font.setBold( bold )
        self.setFont( font )

    def set_font_size( self, size):
        font = self.font()
        font.setPointSize(size)
        self.setFont( font )

    def setIco(self, ico):
        self.setIcon(Ico.icon(ico))

    def setText(self, txt, align=None, ico=None, icon=None, bold=False):
        self.set(txt, align=align, ico=ico, icon=icon, bold=bold)

    def set(self, txt, align=None, ico=None, icon=None, bold=False, font=None, bg=None):

        QtGui.QStandardItem.setText(self, str(txt))
        if align:
            self.setTextAlignment( align)
        if ico:
            self.setIco(ico)
        if icon:
            self.setIcon(icon)

        self.set_bold(bold)

        if font:
            self.set_font_family(font)
        if bg:
            self.set_bg(bg)

    def set_font_family( self, fam):
        font = self.font( )
        font.setFamily( fam )
        self.setFont( font )

    def s(self):
        return str(self.text())

    def i(self):
        x, ok = self.text().toInt()
        if not ok:
            return None
        return x


    def b(self):
        return  str(self.text()) == "1"

    def ds(self):
        return str(self.data().toString())

    def lbl_checked(self,  val, bg_color=None ):
        self.setTextAlignment( QtCore.Qt.AlignCenter)
        if bg_color == None:
            bg_color = "#FFECAA"
        if bool(val):
            self.setText( "Yes")
            self.setData("1")
            self.set_bg( bg_color )
        else:
            self.setText( "-")
            self.setData("")
            self.set_bg( "#efefef")

class ClearButton( QtWidgets.QToolButton ):
    def __init__( self, parent, callback=None ):
        QtWidgets.QToolButton.__init__( self )
        self.setIcon( Ico.icon( Ico.Clear ) )
        self.setToolTip("Clear")
        self.setToolButtonStyle( QtCore.Qt.ToolButtonIconOnly )
        self.setAutoRaise(True)
        self.setFixedWidth(16)

        if callback:
            self.clicked.connect(callback)

class CancelButton( QtWidgets.QPushButton ):
    def __init__( self, parent=None, callback=None ):
        QtWidgets.QPushButton.__init__(self, parent)
        self.setIcon( Ico.icon( Ico.Cancel ) )
        self.setText("Cancel")
        if callback:
            self.clicked.connect(callback)
        else:
            self.clicked.connect(parent.on_reject)


class SaveButton(QtWidgets.QPushButton):
    def __init__(self, parent=None, callback=None, text="Save"):
        QtWidgets.QPushButton.__init__(self, parent)
        self.setIcon(Ico.icon(Ico.Save))
        self.setText(text)
        if callback:
            self.clicked.connect(callback)
        else:
            self.clicked.connect(parent.on_accept)

class IconLabel(QtWidgets.QLabel):

    def __init__( self, parent=None, ico=None):
        QtWidgets.QLabel.__init__( self, parent)

        self.setContentsMargins(5,0,0,0)
        #img_file_path = G.settings.image_path( "/missc/arrow_left_down.gif" )
        icon = Ico.icon(ico)
        self.setPixmap( icon.pixmap(QtCore.QSize( 16, 16 )) )
        self.setFixedWidth(20)




class LNTextEdit(QtWidgets.QFrame):
    """Text widget with support for line numbers
    https://nachtimwald.com/2009/08/19/better-qplaintextedit-with-line-numbers/
    """
    class NumberBar(QtWidgets.QWidget):

        def __init__(self, edit):
            QtWidgets.QWidget.__init__(self, edit)

            self.edit = edit
            self.adjustWidth(1)

        def paintEvent(self, event):
            self.edit.numberbarPaint(self, event)
            QtWidgets.QWidget.paintEvent(self, event)

        def adjustWidth(self, count):
            width = self.fontMetrics().width(unicode(count))
            if self.width() != width:
                self.setFixedWidth(width)

        def updateContents(self, rect, scroll):
            if scroll:
                self.scroll(0, scroll)
            else:
                # It would be nice to do
                # self.update(0, rect.y(), self.width(), rect.height())
                # But we can't because it will not remove the bold on the
                # current line if word wrap is enabled and a new block is
                # selected.
                self.update()

class XPlainTextEdit(QtWidgets.QPlainTextEdit):

    def __init__(self, *args):
        QtWidgets.QPlainTextEdit.__init__(self, *args)

        # self.setFrameStyle(QFrame.NoFrame)

        self.setFrameStyle(QtWidgets.QFrame.NoFrame)
        self.highlight()
        # self.setLineWrapMode(QPlainTextEdit.NoWrap)
        self.setWordWrapMode(QtGui.QTextOption.NoWrap)
        self.setLineWrapMode(QtWidgets.QPlainTextEdit.NoWrap)
        self.setStyleSheet("font-family: monospace")

        self.cursorPositionChanged.connect(self.highlight)

    def highlight(self):
        hi_selection = QtWidgets.QTextEdit.ExtraSelection()

        hi_selection.format.setBackground(self.palette().alternateBase())
        hi_selection.format.setProperty(QtGui.QTextFormat.FullWidthSelection, QtCore.QVariant(True))
        hi_selection.cursor = self.textCursor()
        hi_selection.cursor.clearSelection()

        self.setExtraSelections([hi_selection])

    def numberbarPaint(self, number_bar, event):
        font_metrics = self.fontMetrics()
        current_line = self.document().findBlock(self.textCursor().position()).blockNumber() + 1

        block = self.firstVisibleBlock()
        line_count = block.blockNumber()
        painter = QtWidgets.QPainter(number_bar)
        painter.fillRect(event.rect(), self.palette().base())

        # Iterate over all visible text blocks in the document.
        while block.isValid():
            line_count += 1
            block_top = self.blockBoundingGeometry(block).translated(self.contentOffset()).top()

            # Check if the position of the block is out side of the visible
            # area.
            if not block.isVisible() or block_top >= event.rect().bottom():
                break

            # We want the line number for the selected line to be bold.
            if line_count == current_line:
                font = painter.font()
                font.setBold(True)
                painter.setFont(font)
            else:
                font = painter.font()
                font.setBold(False)
                painter.setFont(font)

            # Draw the line number right justified at the position of the line.
            paint_rect = QtCore.QRect(0, block_top, number_bar.width(), font_metrics.height())
            painter.drawText(paint_rect, Qt.AlignRight, unicode(line_count))

            block = block.next()

        painter.end()


class XLineEdit(QtWidgets.QLineEdit):

    def __init__(self, parent=None):
        QtWidgets.QLineEdit.__init__(self)

    def s(self):
        """return python string"""
        return str(self.text()).strip()

class ToolBarGroup(QtWidgets.QWidget):
    def __init__(self, parent=None, title=None, width=None, hide_labels=False, bg='#999999',
                 is_group=False, toggle_icons=False, toggle_callback=None):
        QtWidgets.QWidget.__init__(self, parent)

        if width:
            self.setFixedWidth(width)

        self.icon_on = Ico.FilterOn
        self.icon_off = Ico.FilterOff
        self.toggle_icons = toggle_icons
        self.toggle_callback = toggle_callback
        self.hide_labels = hide_labels

        self.buttonGroup = None
        self.is_group = is_group
        if self.is_group:
            self.buttonGroup = QtWidgets.QButtonGroup()
            self.buttonGroup.setExclusive(True)
            if self.toggle_callback:
                self.buttonGroup.buttonClicked.connect(self.on_button_clicked)

        self.group_var = None
        self.callback = None
        self.show_icons = True
        self.icon_size = 12
        self.bg_color = bg

        ## Main Layout
        mainLayout = QtWidgets.QVBoxLayout()
        mainLayout.setContentsMargins(0, 0, 0, 0)
        mainLayout.setSpacing(0)
        self.setLayout(mainLayout)

        ## Label
        self.label = QtWidgets.QLabel()
        #bg = "#8F8F8F"  ##eeeeee"
        fg = "#eeeeee"  ##333333"
        lbl_sty = "background: %s; " % self.bg_color  # qlineargradient(x1: 0, y1: 0, x2: 1, y2: 0, stop: 0 #fefefe, stop: 1 #CECECE);"
        lbl_sty += " color: %s; font-size: 8pt; padding: 1px;" % fg  # border: 1px outset #cccccc;"
        self.label.setStyleSheet(lbl_sty)
        self.label.setAlignment(QtCore.Qt.AlignCenter)
        self.label.setFixedHeight(20)
        mainLayout.addWidget(self.label)

        ## Toolbar
        self.toolbar = QtWidgets.QToolBar()
        self.toolbar.setToolButtonStyle(QtCore.Qt.ToolButtonTextBesideIcon)
        self.toolbar.setFixedHeight(30)

        mainLayout.addWidget(self.toolbar)

        if title:
            self.set_title(title)

    def set_title(self, title):
        self.label.setText("%s" % title)

    def addWidget(self, widget):
        self.toolbar.addWidget(widget)
        return widget

    def addAction(self, act):
        self.toolbar.addAction(act)

    def addButton(self, ico=None, text=None, callback=None, idx=None, toggle_callback=None, tooltip=None,
                  ki=None, bold=False, checkable=False, checked=None, width=None, return_action=False):

        butt = XToolButton()

        if self.is_group:
            if idx != None:
                self.buttonGroup.addButton(butt, idx)
            else:
                self.buttonGroup.addButton(butt)
        if self.hide_labels == False:
            if text != None:
                butt.setText(text)
        if text == None:
            butt.setToolButtonStyle(QtCore.Qt.ToolButtonIconOnly)
        else:
            butt.setToolButtonStyle(QtCore.Qt.ToolButtonTextBesideIcon)
        if tooltip:
            butt.setToolTip(tooltip)
        if self.toggle_icons:
            butt.setIconSize(QtCore.QSize(10, 10))
            butt.setIcon(Ico.icon(self.icon_off))
        if ico:
            butt.setIcon(Ico.icon(ico))
            butt.setIconSize(QtCore.QSize(10, 10))

        butt.setCheckable(checkable)
        if checked != None:
            butt.setChecked(checked)


        butt.setProperty("ki", ki)
        nuAct = self.toolbar.addWidget(butt)
        if callback:
            self.connect(butt, QtCore.SIGNAL("clicked()"), callback)
        #if toggle_callback:
        #   self.connect(butt, QtCore.SIGNAL("toggled(bool)"), toggle_callback)
        if bold:
            self.set_bold(butt)
        if width:
            butt.setFixedWidth(width)

        self.on_button_clicked(block=True)

        if return_action:
            return nuAct
        return butt

    def set_bold(self, w):
        f = w.font()
        f.setBold(True)
        w.setFont(f)

    def on_button_clicked(self, butt=None, block=False):
        if self.is_group:
            for b in self.buttonGroup.buttons():
                b.setIcon( Ico.icon(self.icon_on if b.isChecked() else self.icon_off) )

                if block == False and b.isChecked():
                    if self.toggle_callback:
                        self.toggle_callback(self.buttonGroup.id(b))

    def get_id(self):
        id = self.buttonGroup.checkedId()
        if id == -1:
            return None
        return id


class TopStatusBar(QtWidgets.QWidget):
    """A QWidget with many embedded widgets for StatusBar"""

    sigRefresh = pyqtSignal()

    def __init__(self, parent=None, refresh=True, statusLabel=False,
                 pager=False, modeLabel=True):
        QtWidgets.QWidget.__init__(self, parent)


        self.refresh = refresh
        self.includeStatusLabel = statusLabel

        self.mainLayout = QtWidgets.QHBoxLayout()
        m = 0
        self.mainLayout.setContentsMargins(m, m, m, m)
        self.setLayout(self.mainLayout)

        ##== Status Bar
        self.statusBar = QtWidgets.QStatusBar()
        self.statusBar.setSizeGripEnabled(False)
        self.showMessage(" ")
        self.mainLayout.addWidget(self.statusBar, 1)


        ##== Progress Bar
        progressWidget = QtWidgets.QWidget()
        progressWidget.setMinimumWidth(40)
        progressWidget.setMaximumWidth(200)
        self.mainLayout.addWidget(progressWidget)

        progressLayout = QtWidgets.QHBoxLayout()
        m = 0
        progressLayout.setContentsMargins(m, m, m, m)
        progressWidget.setLayout(progressLayout)

        self.progress = QtWidgets.QProgressBar(self)
        # self.progress.setFixedWidth( 40 )
        # self.progress.setFixedHeight( 10 )
        self.setContentsMargins(m, m, m, m)
        self.progress.setMinimum(0)
        self.progress.setMaximum(1)
        self.progress.setInvertedAppearance(False)
        self.progress.setTextVisible(False)
        self.progress.setVisible(False)
        # self.connect( self.progress, QtCore.SIGNAL('valueChanged(int)'), self.on_progress_value_changed)
        progressLayout.addWidget(self.progress, 1)

        ################################################################
        ### Status Label
        if self.includeStatusLabel:
            wid = 40
            self.lblStatus = QtWidgets.QLabel("Status:Label")
            self.lblStatus.setMinimumWidth(wid)
            self.lblStatus.setMaximumWidth(wid)
            self.lblStatus.setStyleSheet("font-size: 8pt")
            self.mainLayout.addWidget(self.lblStatus, 1)


        ################################################################
        ## Refrehsing Button
        """
        if self.refresh:
            self.buttRefresh = QtWidgets.QToolButton()
            self.mainLayout.addWidget(self.buttRefresh)

            self.buttRefresh.setIcon(Ico.icon(Ico.Refresh))
            # self.buttRefresh.setFlat(True)
            self.buttRefresh.setAutoRaise(True)
            self.buttRefresh.setIconSize(QtCore.QSize(16, 16))
            self.buttRefresh.setStyleSheet("padding: 0px;")
            self.connect(self.buttRefresh, QtCore.SIGNAL("clicked()"), self.on_refresh)
            if hasattr(parent, "on_refresh"):
                self.connect(self.buttRefresh, QtCore.SIGNAL("clicked()"), parent.on_refresh)

            # self.buttRefresh.setFixedWidth(40)

            self.popMenu = QtWidgets.QMenu()
            self.actionDebug = self.popMenu.addAction("Debug", self.on_server_debug_button)
            self.actionDebug.setIcon(dIco.icon(dIco.Dev))
            self.actionDebug.setIconVisibleInMenu(True)
            self.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
            self.connect(self, QtCore.SIGNAL("customContextMenuRequested(QPoint)"), self.on_menu)
        else:
            self.popMenu = None
        """

    def showMessage( self, mess, timeout=None, warn=None, info=None ):
        if warn:
            self.statusBar.setStyleSheet("color: #990000;")
        elif info:
            self.statusBar.setStyleSheet("color: #3361C2;")
        else:
            self.statusBar.setStyleSheet("")

        if timeout == None:
            self.statusBar.showMessage( mess )
        else:
            self.statusBar.showMessage( mess, timeout )



# -*- coding: utf-8 -*-

import os
from Qt import QtGui, QtWidgets, QtCore, Qt, pyqtSignal

import app_globals as G

from ogt import ags4, FORMATS

from img import Ico
import xwidgets
import ags4_models
import ogtgui_group




class AbbrPickDialog( QtWidgets.QDialog ):

    def __init__( self, parent=None, ogtHead=None, editCell=None):
        QtWidgets.QDialog.__init__( self, parent )

        self.ogtHead = ogtHead
        self.editCell = editCell

        self.selected = None


        self.mainLayout = xwidgets.vlayout(margin=0)
        self.setLayout(self.mainLayout)

        self.setWindowTitle("Picklist")
        self.setWindowIcon(Ico.icon(Ico.TypePicklist))

        topLay = xwidgets.hlayout(spacing=5)
        self.mainLayout.addLayout(topLay)

        sty = "background-color: #555555; color: #dddddd; padding: 4px;"
        self.lblTitle = xwidgets.XLabel(text="-", style=sty)
        topLay.addWidget(self.lblTitle)

        self.mainLayout.addSpacing(10)

        self.modelAbbrs = ags4_models.AbbrevItemsModel(ogtHead=self.ogtHead)
        #self.modelAbbrs.sigCheckedChanged.connect(self.on_checked_changed)

        #== tree
        self.tree = QtWidgets.QTreeView()
        self.mainLayout.addWidget(self.tree)
        self.tree.setModel(self.modelAbbrs)

        self.tree.setExpandsOnDoubleClick(False)
        self.tree.setRootIsDecorated(False)
        self.tree.setUniformRowHeights(True)
        self.tree.setAlternatingRowColors(True)
        self.tree.setMinimumHeight(700)
        self.tree.setMinimumWidth(500)

        #self.tree.setSelectionMode(QtWidgets.QAbstractItemView.SelectRows)
        #self.tree.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
        self.tree.setSelectionMode(QtWidgets.QAbstractItemView.NoSelection)
        #self.tree.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)

        CA = ags4_models.AbbrevItemsModel.CA

        self.tree.header().setSectionResizeMode(CA.description, QtWidgets.QHeaderView.Stretch)

        self.tree.setColumnWidth(CA.check, 15)
        self.tree.setColumnWidth(CA.code, 80)
        #self.tree.setColumnWidth(ags4_models.AbbrevItemsModel.CA.description, 200)

        self.statusBar = QtWidgets.QStatusBar()
        self.mainLayout.addWidget(self.statusBar)


        ## ================ Button BAr
        buttLay = xwidgets.hlayout(spacing=5, margin=5)
        self.mainLayout.addLayout(buttLay)

        self.buttCancel = xwidgets.XPushButton(callback=self.on_add, ico=Ico.Add, text="Add")
        buttLay.addWidget(self.buttCancel)


        lbl = xwidgets.XLabel(text="Right click to edit")
        buttLay.addWidget(lbl)
        buttLay.addStretch(3)

        self.buttCancel = xwidgets.CancelButton(callback=self.on_cancel)
        buttLay.addWidget(self.buttCancel)

        self.buttSave = xwidgets.SaveButton(callback=self.on_save, text="Select")
        buttLay.addWidget(self.buttSave)

        #self.modelAbbrs.set_checked(editCell.value)
        self.do_select_update()

    def on_checked_changed(self):
        self.update_statusBar()

    def on_add(self):
        self.show_edit_dialog(dict(code="", description="", list="custom"))

    def on_edit(self):
        pass

    def show_edit_dialog(self, rec):
        d = AbbrEditDialog(rec=rec)
        if d.exec_():
            self.ogtHead.parentGroup.parentDoc.add_abbr(self.ogtHead.head_code, d.rec)
            print(d.rec, self)
            self.modelAbbrs.load_data()
            self.modelAbbrs.set_checked(d.rec)

    def do_select_update(self):

        sels = []
        if self.ogtHead.parent_group.parentDoc.concat_char() in self.editCell.value:
            sels = self.editCell.value.split(self.ogtHead.parentGroup.parentDoc.concat_char())
        else:
            sels = self.editCell.value

        self.modelAbbrs.set_checked(sels)

        s = "<b>%s</b>: %s" % (self.ogtHead.head_code, self.ogtHead.head_description)
        self.lblTitle.setText(s)

        self.update_statusBar()

    def update_statusBar(self):
        codes = self.modelAbbrs.checked_codes()
        #lst = ["<b>%s</b>" % c for c in codes]
        sta = "%s - %s selected [%s]" % (self.modelAbbrs.rowCount(), len(codes), ",".join(codes) )
        self.statusBar.showMessage(sta)


    def on_cancel(self):
        self.reject()

    def on_save(self):

        rec = self.modelAbbrs.get_checked()
        if rec is not None:
            self.editCell.value = rec['code']
        self.ogtHead.emit()
        self.accept()



    # def sson_tree_clicked(self, midx):
    #     val = self.model.val_from_midx(midx)
    #     self.cell.value = val['code']
    #     self.accept()



class AbbrEditDialog( QtWidgets.QDialog ):

    def __init__( self, parent=None, rec=None):
        QtWidgets.QDialog.__init__( self, parent )

        self.rec = rec

        self.setWindowTitle("Edit Abbrev")
        self.setWindowIcon(Ico.icon(Ico.AgsAbbrevItem))
        self.setMinimumWidth(600)

        self.mainLayout = xwidgets.vlayout(margin=10, spacing=10)
        self.setLayout(self.mainLayout)


        self.grid = QtWidgets.QGridLayout()
        self.mainLayout.addLayout(self.grid)

        self.grid.addWidget(QtWidgets.QLabel("Code [ABBR_CODE]"), 0, 0)
        self.grid.addWidget(QtWidgets.QLabel("Description  [ABBR_DESC]"), 0, 1)
        self.grid.addWidget(QtWidgets.QLabel("List  [ABBR_LIST]"), 0, 2)

        self.txtCode = xwidgets.XLineEdit()
        self.grid.addWidget(self.txtCode, 1, 0)

        self.txtDesc = xwidgets.XLineEdit()
        self.grid.addWidget(self.txtDesc, 1, 1)

        self.txtlist = xwidgets.XLineEdit()
        self.grid.addWidget(self.txtlist, 1, 2)


        self.grid.setColumnStretch(0, 1)
        self.grid.setColumnStretch(1, 2)
        self.grid.setColumnStretch(2, 1)

        ## ================
        buttLay = xwidgets.hlayout()
        self.mainLayout.addLayout(buttLay)

        buttLay.addStretch(3)

        self.buttCancel = xwidgets.CancelButton(callback=self.on_cancel)
        buttLay.addWidget(self.buttCancel)

        self.buttSave = xwidgets.SaveButton(callback=self.on_save, text="Save")
        buttLay.addWidget(self.buttSave)

        self.txtCode.editingFinished.connect(self.on_code_edited)
        self.load_rec()


    def load_rec(self):
        if self.rec == None:
            return
        self.txtCode.setText(self.rec['code'])
        self.txtDesc.setText(self.rec['description'])
        self.txtlist.setText(self.rec['list'])

    def on_code_edited(self):
        self.txtCode.setText(self.txtCode.text().upper())

    def on_cancel(self):
        self.reject()

    def on_save(self):
        self.rec = dict(code=self.txtCode.s(), description=self.txtDesc.s(), list=self.txtlist.s())
        self.accept()



class DataTypeSelectDialog( QtWidgets.QDialog ):

    def __init__( self, parent, ogtHead):
        QtWidgets.QDialog.__init__( self, parent )

        self.ogtHead = ogtHead

        self.mainLayout = xwidgets.vlayout()
        self.setLayout(self.mainLayout)

        self.setWindowTitle("Select Type")
        self.setWindowIcon(Ico.icon(Ico.TypePicklist))


        self.tree = QtWidgets.QTreeView()
        self.tree.setRootIsDecorated(False)
        self.tree.setAlternatingRowColors(True)
        self.tree.setMinimumHeight(600)
        self.tree.setMinimumWidth(400)
        self.mainLayout.addWidget(self.tree)

        self.model = G.ags.modelDataTypes
        self.tree.setModel(self.model)
        self.tree.clicked.connect(self.on_tree_clicked)

        idx = self.model.midx_from_val(  self.ogtHead.data_type_cell.value  )
        if idx:
            self.tree.selectionModel().select(idx, QtCore.QItemSelectionModel.SelectCurrent|QtCore.QItemSelectionModel.Rows)

    def on_tree_clicked(self, midx):
        val = G.ags.modelDataTypes.val_from_midx(midx)
        self.ogtHead.data_type_cell.value = val['code']
        self.accept()


class UnitSelectDialog( QtWidgets.QDialog ):


    def __init__( self, parent, ogtHead):
        QtWidgets.QDialog.__init__( self, parent )

        self.ogtHead = ogtHead

        self.mainLayout = xwidgets.vlayout()
        self.setLayout(self.mainLayout)

        self.setWindowTitle("Select Unit")
        self.setWindowIcon(Ico.icon(Ico.TypePicklist))


        self.tree = QtWidgets.QTreeView()
        self.tree.setRootIsDecorated(False)
        self.tree.setAlternatingRowColors(True)
        self.tree.setMinimumHeight(600)
        self.mainLayout.addWidget(self.tree)

        self.tree.setModel(G.ags.modelUnits)
        self.tree.clicked.connect(self.on_tree_clicked)

        idx = G.ags.modelUnits.midx_from_val(  self.ogtHead.unit_cell.value  )
        if idx:
            self.tree.selectionModel().select(idx, QtWidgets.QItemSelectionModel.SelectCurrent|QtWidgets.QItemSelectionModel.Rows)

    def on_tree_clicked(self, midx):
        val = G.ags.modelUnits.val_from_midx(midx)
        self.ogtHead.unit_cell.value = val['unit']
        self.accept()



class PickIDDialog( QtWidgets.QDialog ):

    def __init__( self, parent, ogtGroup, ogtHead, cell):
        QtWidgets.QDialog.__init__( self, parent )

        self.ogtGroup = ogtGroup
        self.ogtHead = ogtHead
        self.cell = cell


        self.mainLayout = xwidgets.vlayout()
        self.setLayout(self.mainLayout)

        self.setWindowTitle("Pick ID")
        self.setWindowIcon(Ico.icon(Ico.TypePicklist))


        self.tree = QtWidgets.QTreeView()
        self.mainLayout.addWidget(self.tree)


        self.tree.setRootIsDecorated(False)
        self.tree.setUniformRowHeights(True)
        self.tree.setAlternatingRowColors(True)
        self.tree.setMinimumHeight(700)
        self.tree.setMinimumWidth(900)

        self.tree.setSelectionMode(QtWidgets.QAbstractItemView.SingleSelection)
        self.tree.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)

        self.model = ogtgui_group.GroupDataModel(self, self.ogtGroup)
        self.tree.setModel(self.model)

        self.tree.clicked.connect(self.on_tree_clicked)

        self.do_select()

    def do_select(self):
        cidx = self.ogtGroup.index_of_heading(self.ogtHead.head_code)
        if self.ogtGroup.ogtDoc.concat_char in self.cell.value:
            dsa

        idx = self.model.midx_from_val(cidx, self.cell.value)
        if idx:
            self.tree.selectionModel().select(idx,
                                              QtWidgets.QItemSelectionModel.SelectCurrent | QtWidgets.QItemSelectionModel.Rows)


    def on_tree_clicked(self, midx):
        xcell = self.model.val_from_midx(midx)
        self.cell.value = xcell.value
        self.accept()


class HeadingsSelectDialog( QtWidgets.QDialog ):


    def __init__( self, parent, ogtGroup):
        QtWidgets.QDialog.__init__( self, parent )

        self.ogtGroup = ogtGroup

        self.mainLayout = xwidgets.vlayout()
        self.setLayout(self.mainLayout)

        self.setWindowTitle("Select Headings")
        self.setWindowIcon(Ico.icon(Ico.TypePicklist))
        self.setMinimumWidth(700)
        self.setMinimumHeight(700)

        self.model = ags4_models.HeadingsModel(self)


        self.tree = QtWidgets.QTreeView()
        self.tree.setModel(self.model)
        self.mainLayout.addWidget(self.tree)

        self.tree.setRootIsDecorated(False)
        self.tree.setAlternatingRowColors(True)
        self.tree.setMinimumHeight(600)

        self.tree.setSelectionMode(QtWidgets.QAbstractItemView.NoSelection)

        self.tree.clicked.connect(self.on_tree_clicked)

        self.model.set_group(ags4.DD.group(self.ogtGroup.group_code))


        self.buttGroup = QtWidgets.QButtonGroup()
        self.buttGroup.setExclusive(False)
        self.buttGroup.buttonClicked.connect(self.on_add_button)

        self.setup_buttons()
        self.update_buttons()

        self.tree.update()

    def setup_buttons(self):

        for ridx in range(0, self.model.rowCount()):
            midx = self.model.index(ridx, 0)
            butt = xwidgets.XToolButton()
            self.tree.setIndexWidget(midx, butt)
            self.buttGroup.addButton(butt, ridx)

    def update_buttons(self):

        for ridx in range(0, self.model.rowCount()):

            midx = self.model.index(ridx, 0)
            butt = self.tree.indexWidget(midx)

            heading = self.model.data(self.model.index(ridx, ags4_models.HeadingsModel.CH.head_code), Qt.DisplayRole)
            exists = heading in self.ogtGroup.headings_index

            if exists:
                butt.setText("Remove")
                butt.setIco(Ico.Minus)
                butt.setVisible(False)
            else:
                butt.setText("Add")
                butt.setIco(Ico.Plus)
                butt.show()
                butt.setVisible(True)




    def on_add_button(self, butt):
        ridx = self.buttGroup.id(butt)

        midx = self.model.index(ridx, ags4_models.HeadingsModel.CH.head_code)
        rec = self.model.rec_from_midx(midx)

        ogtHead = self.ogtGroup.add_heading(rec)
        print("on_add", ogtHead, self)
        self.update_buttons()


    def on_tree_clicked(self, midx):
        print(midx)

        return
        val = G.ags.modelUnits.val_from_midx(midx)
        self.ogtHead.unit_cell.value = val['unit']
        self.accept()


class EXT:
    ags = ["ags", "ags4"]
    excel = ["xlsx"]
    zip = ["zip"]
    all = ags + excel + zip

class XFileSystemModel(QtWidgets.QFileSystemModel):

    def __init__( self, parent=None):
        QtWidgets.QFileSystemModel.__init__( self, parent )

    def _get_ext(self, midx):
        return QtCore.QFileInfo( self.data(midx, Qt.DisplayRole).lower() ).suffix()

    def data(self, midx, role ):

        if midx.column() == 0:
            if role == Qt.DecorationRole:

                ext = self._get_ext(midx)

                if not self.isDir(midx):


                    if ext in EXT.ags:
                        return Ico.icon(Ico.Ags4)

                    if ext in EXT.excel:
                        return Ico.icon(Ico.Excel)

                    #if ext in EXT.zip:
                    #    return Ico.icon(Ico.Zip)

            if role == Qt.FontRole:

                if self._get_ext(midx) in EXT.all:
                    f = QtGui.QFont()
                    f.setBold(True)
                    return f


        return QtWidgets.QFileSystemModel.data(self, midx, role)

class FileSelectDialog( QtWidgets.QDialog ):

    def __init__( self, parent, multi=False, root=None):
        QtWidgets.QDialog.__init__( self, parent )

        self.multi = multi

        self.file_path = None

        self.mainLayout = xwidgets.vlayout()
        self.setLayout(self.mainLayout)

        self.setWindowTitle("Select Files")
        self.setWindowIcon(Ico.icon(Ico.TypePicklist))

        ##= Model
        #ROOT_PATH = "/home/ogt"
        self.model = XFileSystemModel()



        ##= Tree
        self.tree = QtWidgets.QTreeView()
        self.tree.setModel(self.model)
        self.mainLayout.addWidget(self.tree)


        #self.tree.setRootIsDecorated(False)
        self.tree.setUniformRowHeights(True)
        self.tree.setAlternatingRowColors(True)
        self.tree.setMinimumHeight(700)
        self.tree.setMinimumWidth(900)

        self.tree.setColumnWidth(0, 250)


        self.tree.setSelectionMode(QtWidgets.QAbstractItemView.MultiSelection if multi else QtWidgets.QAbstractItemView.SingleSelection)
        self.tree.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)

        self.tree.clicked.connect(self.on_tree_clicked)

        buttlay = xwidgets.hlayout(margin=3)
        self.mainLayout.addLayout(buttlay)
        buttlay.addStretch(12)

        self.buttCancel = xwidgets.CancelButton(self)
        buttlay.addWidget(self.buttCancel)

        self.buttSave = xwidgets.SaveButton(self, text="Open")
        buttlay.addWidget(self.buttSave)

        self.set_root(root)

    def set_root(self, root):
        if root == None:
            #root = self.model.myComputer()
            root = os.path.expanduser("~")
        self.model.setRootPath(root)
        self.tree.setRootIndex(self.model.index(root))


    def on_reject(self):
        self.reject()

    def on_accept(self):

        midx = self.tree.selectedIndexes()[0]
        self.file_path = self.model.data(midx, QtWidgets.QFileSystemModel.FilePathRole)

        self.accept()



    def on_tree_clicked(self, midx):
        pass


class SaveExportDialog( QtWidgets.QDialog ):

    def __init__( self, parent, ogtDoc):
        QtWidgets.QDialog.__init__( self, parent )

        self.ogtDoc = ogtDoc

        self.setWindowTitle("Save & Export Files")
        self.setWindowIcon(Ico.icon(Ico.Export))#
        self.setMinimumWidth(700)
        self.setMinimumHeight(600)

        self.mainLayout = xwidgets.vlayout(margin=10)
        self.setLayout(self.mainLayout)

        self.buttGroup = QtWidgets.QButtonGroup()
        self.buttGroup.setExclusive(False)
        self.buttGroup.buttonClicked.connect(self.on_button)

        self.buttGroupFormat = QtWidgets.QButtonGroup()
        self.buttGroupFormat.setExclusive(True)
        self.buttGroupFormat.buttonClicked.connect(self.on_button_preview_format)

        self.grid = QtWidgets.QGridLayout()
        self.grid.setSpacing(5)
        self.mainLayout.addLayout(self.grid)

        sty = "background-color: white; padding: 3px;"
        row = 0
        self.grid.addWidget(xwidgets.XLabel(text="File Path:"), row, 0, 1, 1, Qt.AlignRight)
        self.lblFilePath = xwidgets.XLabel(text=os.path.dirname(self.ogtDoc.source_file_path),
                                           style=sty)
        self.grid.addWidget(self.lblFilePath, row, 1)

        row += 1
        self.grid.addWidget(xwidgets.XLabel(text="File Name:"), row, 0, 1, 1, Qt.AlignRight)
        self.lblFileName = xwidgets.XLabel(text=os.path.basename(self.ogtDoc.source_file_path),
                                           style=sty)
        self.grid.addWidget(self.lblFileName, row, 1)

        row += 1
        self.grid.addWidget(xwidgets.XLabel(text="Format:"), row, 0, 1, 1, Qt.AlignRight|Qt.AlignTop)

        self.lbls = []
        radLay = QtWidgets.QGridLayout()
        radLay.setSpacing(5)
        self.grid.addLayout(radLay, row, 1)
        for xidx, ext in enumerate(["json", "yaml", "xlsx", "ags4"]):
            butt = QtWidgets.QCheckBox()
            butt.setText(ext)
            radLay.addWidget(butt, xidx, 0)
            self.buttGroup.addButton(butt, xidx)

            bn = os.path.basename(self.ogtDoc.source_file_path) + "." + ext
            lbl = xwidgets.XLabel(text=bn)
            radLay.addWidget(lbl, xidx, 1)
            lbl.hide()
            self.lbls.append(lbl)
        radLay.setColumnStretch(0, 1)
        radLay.setColumnStretch(1, 4)

        row += 1
        self.grid.addWidget(xwidgets.XLabel(text="Extended:"), row, 0, 1, 1, Qt.AlignRight | Qt.AlignTop)
        self.chkExtended = QtWidgets.QCheckBox()
        self.chkExtended.setText("include descriptions etc")
        self.grid.addWidget(self.chkExtended, row, 1)
        self.buttGroup.addButton(self.chkExtended)

        row += 1
        self.grid.addWidget(xwidgets.XLabel(text="Minify:"), row, 0, 1, 1, Qt.AlignRight | Qt.AlignTop)
        self.chkMinify = QtWidgets.QCheckBox()
        self.chkMinify.setText("Removed whitespace (json only)")
        self.grid.addWidget(self.chkMinify, row, 1)
        self.buttGroup.addButton(self.chkMinify)

        row += 1
        self.grid.addWidget(xwidgets.XLabel(text="Overwrite:"), row, 0, 1, 1, Qt.AlignRight | Qt.AlignTop)
        self.chkOverwrite = QtWidgets.QCheckBox()
        self.chkOverwrite.setText("Overwrite existing files")
        self.grid.addWidget(self.chkOverwrite, row, 1)


        row += 1
        #self.grid.addWidget(xwidgets.XLabel(text="Preview:"), row, 0, Qt.AlignRight | Qt.AlignTop)

        radLay = xwidgets.hlayout(spacing=5)
        radLay.addWidget(xwidgets.XLabel(text="Preview:"), 0)

        self.grid.addLayout(radLay, row, 0, 1, 2,  Qt.AlignLeft)
        for idx, ext in enumerate(["json", "yaml", "ags4"]):
            butt = QtWidgets.QRadioButton()
            butt.setText(ext)
            radLay.addWidget(butt, 0)
            if idx == 0:
                butt.setChecked(True)
            self.buttGroupFormat.addButton(butt, idx)
        radLay.addStretch(10)

        row += 1
        self.txtPreview = QtWidgets.QTextEdit()
        self.txtPreview.setStyleSheet("font-size: 8pt; font-family: monospace;")
        self.txtPreview.setMinimumHeight(300)
        self.grid.addWidget(self.txtPreview, row, 0, 1, 2)

        self.grid.setColumnStretch(0, 1)
        self.grid.setColumnStretch(1, 3)

        self.mainLayout.addSpacing(10)

        self.botLay = xwidgets.hlayout()
        self.mainLayout.addLayout(self.botLay)

        self.botLay.addStretch(10)

        self.statusBar = QtWidgets.QStatusBar()
        self.botLay.addWidget(self.statusBar, 10)
        #self.buttCancel = xwidgets.CancelButton(self)
        #self.botLay.addWidget(self.buttCancel)

        self.buttSave = xwidgets.XPushButton(text="Write Files", ico=Ico.Export, callback=self.on_accept)

        self.botLay.addWidget(self.buttSave)
        self.buttSave.setEnabled(False)

        self.update_preview()

    def on_button(self, butt=None):

        can_save = False
        for id in range(0, 4):
            butt = self.buttGroup.button(id)
            #print(butt.text(), butt.isChecked())
            self.lbls[id].setVisible(butt.isChecked())
            if butt.isChecked():
                can_save = True

        self.buttSave.setEnabled(can_save)
        self.update_preview()

    def on_button_preview_format(self):
        self.update_preview()

    def update_preview(self):

        self.ogtDoc.opts.minify = self.chkMinify.isChecked()
        self.ogtDoc.opts.extended = self.chkExtended.isChecked()

        if self.buttGroupFormat.checkedId() == 0:
            s, err = self.ogtDoc.to_json()

        elif self.buttGroupFormat.checkedId() == 1:
            s, err = self.ogtDoc.to_yaml()
        else:
            s, err = self.ogtDoc.to_ags4()
        self.txtPreview.setText(s)

    def on_reject(self):
        self.reject()

    def on_accept(self):

        over = True
        sty_green = "color: green;"
        ## json
        if self.buttGroup.button(0).isChecked():
            #fp = self.ogtDoc.source_file_path + ".json"
            mess, err = self.ogtDoc.write(ext="json", beside=True, overwrite=over, file_path=None)
            print(mess, err)
            if err is None:
                self.lbls[0].setStyleSheet(sty_green)


        ## yaml
        if self.buttGroup.button(1).isChecked():
            #fp = self.ogtDoc.source_file_path + ".yaml"
            mess, err = self.ogtDoc.write(ext="yaml", beside=True, overwrite=over, file_path=None)
            print(mess, err)
            if err is None:
                self.lbls[1].setStyleSheet(sty_green)

        ## xlsx
        if self.buttGroup.button(2).isChecked():
            #fp = self.ogtDoc.source_file_path + ".xlsx"
            mess, err = self.ogtDoc.write(ext="xlsx", beside=True, overwrite=over, file_path=None)
            print(mess, err)
            if err is None:
                self.lbls[2].setStyleSheet(sty_green)

        self.statusBar.showMessage("Saved", 2000)





# -*- coding: utf-8 -*-


from Qt import QtGui, QtWidgets, QtCore, Qt, pyqtSignal
from ogt import CELL_COLORS
import app_globals as G

import ogtgui_group
import ogtgui_widgets
import ags4_widgets
import xwidgets
from img import Ico


class OGTDocumentWidget( QtWidgets.QWidget ):

    sigUpdatedXX = pyqtSignal(object)
    #sigChanged = pyqtSignal()


    def __init__( self, parent=None, ogtDoc=None):
        QtWidgets.QWidget.__init__( self, parent )

        self.ogtDoc = ogtDoc

        self.show_data_count = False
        self.show_description = False
        self._tab_widgets = [] # dodgy

        self.mainLayout = QtWidgets.QVBoxLayout()
        self.mainLayout.setSpacing(0)
        self.mainLayout.setContentsMargins(0,0,0,0)
        self.setLayout(self.mainLayout)

        self.headerWidget = ogtgui_widgets.HeaderBarWidget(title="Groups", bg="#88EEFF")
        self.mainLayout.addWidget(self.headerWidget, 0)

        # == View Mode
        self.tbgView = xwidgets.ToolBarGroup(title="Groups Sort",
                                             is_group=True, toggle_icons=True,
                                             toggle_callback=self.on_butt_sort)
        self.headerWidget.addWidget(self.tbgView)

        for idx, optt in enumerate([[self.ogtDoc.SORT.recommended, "Recommended"],
                                    [self.ogtDoc.SORT.a2z, "Alphabetic"],
                                    [self.ogtDoc.SORT.source, "Sources"]]):
            #buttSort = xwidgets.XToolButton(text=optt[1], checkable=True, checked=idx == 0)

            self.tbgView.addButton(text=optt[1], idx=optt[0], checkable=True, checked=idx == 0)

        self.chkShowDescription = QtWidgets.QCheckBox()
        self.chkShowDescription.setText("Show Description")
        self.chkShowDescription.setChecked(True)
        self.headerWidget.addWidget(self.chkShowDescription)
        self.chkShowDescription.toggled.connect(self.set_show_description)


        ##== Scroll aread
        self.scrollArea = QtWidgets.QScrollArea()
        self.scrollArea.setWidgetResizable(True)
        self.scrollArea.setFixedHeight(70)
        self.mainLayout.addWidget(self.scrollArea, 0)

        self.scrollWidget = QtWidgets.QWidget()
        #self.scrollWidget.setFixedHeight(80)
        self.scrollArea.setWidget(self.scrollWidget)

        self.scrollLayout = QtWidgets.QHBoxLayout()
        self.scrollLayout.setContentsMargins(0, 0, 0, 0)
        self.scrollLayout.setSpacing(0)
        self.scrollWidget.setLayout(self.scrollLayout)
        self.scrollLayout.addStretch(100)

        self.stackWidget = QtWidgets.QStackedWidget()
        self.mainLayout.addWidget(self.stackWidget)

        self.scrollArea.horizontalScrollBar().rangeChanged.connect(self.on_detect_scroll)


    def init_load(self):
        pass


    def do_update(self):

        for grpOb in self.ogtDoc.groups_list:
            widget = self.add_set_group( grpOb )


        for idx in range(0, self.stackWidget.count()):
            self.stackWidget.widget(idx).do_update()

        self.update_group_tabs()
        print("so_iupdate", self)


        self.on_detect_scroll()

        cidx = self.stackWidget.currentIndex()
        if cidx == -1:
            return
        gc = self.stackWidget.currentWidget().ogtGroup.group_code
        self.select_group(gc)

        #if G.args.dev:
        #    self.select_group("LOCA")



    def on_butt_sort(self, sort_id):
        self.ogtDoc.set_sort_mode(sort_id)
        print("on_butt_sort", sort_id, self)
        self.do_update()

    def on_detect_scroll(self):
        viz = self.scrollArea.horizontalScrollBar().isVisible()
        self.scrollArea.setFixedHeight(90 if viz else 70)
        # self.scrollWidget.update()

    def set_show_description(self, xstate=None):
        state = self.chkShowDescription.isChecked()
        for idx in range(0, self.stackWidget.count()):
            self.stackWidget.widget(idx).set_show_description(state)

    def add_set_group(self, ogtGroup):

        # first check this not already there
        for idx in range(0, self.stackWidget.count()):
            if self.stackWidget.widget(idx).ogtGroup.group_code == ogtGroup.group_code:
                return self.stackWidget.widget(idx)


        grpViewWidget = ogtgui_group.GroupViewWidget(ogtGroup=ogtGroup)
        grpViewWidget.do_update()
        idx = self.stackWidget.addWidget(grpViewWidget)

        grpViewWidget.sigGoto.connect(self.on_goto)
        #self.ogtDoc.sigChanged.connect(self.on_changed)
        #grpViewWidget.sigChanged.connect(self.on_changed)

        return grpViewWidget

    def update_group_tabs(self):


        curr_grp_code = None
        if self.stackWidget.currentWidget():
            curr_grp_code = self.stackWidget.currentWidget().ogtGroup.group_code



        ## nuke existsing buttons
        while self.scrollLayout.count() > 0:
            w = self.scrollLayout.takeAt(0)
            w = None


        self._tab_widgets = []

        for idx, ogtGroup in enumerate(self.ogtDoc.groups_list):

            gWid = XTabWidget(ogtGroup)
            gWid.setFixedWidth(55)
            self.scrollLayout.addWidget(gWid)
            self._tab_widgets.append(gWid)
            gWid.sigSelected.connect(self.on_grp_tab_clicked)

        self.scrollLayout.addStretch(10)

        self.select_group(curr_grp_code)


    #
    # def on_changed_DEAD(self):
    #     print("on_changed", self)
    #     self.sigChanged.emit()


    def select_group(self, group_code):

        if self.stackWidget.currentWidget():
            if group_code != self.stackWidget.currentWidget().ogtGroup.group_code:
                for i in range(0, self.stackWidget.count()):
                    if self.stackWidget.widget(i).ogtGroup.group_code == group_code:
                        self.stackWidget.setCurrentIndex(i)
                        break

        for tw in self._tab_widgets:
            tw.set_selected(tw.ogtGroup.group_code == group_code)





    def on_goto(self, xcode):
        if "_" in xcode:
            grp_code = xcode.split("_")[0]
        else:
            grp_code = xcode
        self.select_group(grp_code)


    def on_grp_tab_clicked(self, grp_code):
        self.select_group(grp_code)



class XTabWidget(QtWidgets.QWidget):

    sigSelected = pyqtSignal(str)

    def __init__(self, ogtGroup):
        QtWidgets.QWidget.__init__(self)

        self.ogtGroup = ogtGroup

        self.selected = False

        self.W = 2

        m = 0
        lay = QtWidgets.QVBoxLayout()
        lay.setContentsMargins(m,m,m,m)
        lay.setSpacing(0)
        self.setLayout(lay)

        ## Coloured red when required
        self.lblBarTop = QtWidgets.QLabel()
        self.lblBarTop.setFixedHeight(4)
        lay.addWidget(self.lblBarTop)

        ## Group code label
        self.lblGroupCode = xwidgets.XLabel(bold=True)
        self.lblGroupCode.setAlignment(Qt.AlignCenter)
        lay.addWidget(self.lblGroupCode)
        self.lblGroupCode.sigClicked.connect(self.on_lbl_clicked)

        # Errors Count label
        self.lblErrorCount = xwidgets.XLabel()
        self.lblErrorCount.setAlignment(Qt.AlignCenter)
        lay.addWidget(self.lblErrorCount)
        self.lblErrorCount.sigClicked.connect(self.on_lbl_clicked)

        # data count label
        self.lblDataRowsCount = xwidgets.XLabel()
        self.lblDataRowsCount.setAlignment(Qt.AlignCenter)
        lay.addWidget(self.lblDataRowsCount)
        self.lblDataRowsCount.sigClicked.connect(self.on_lbl_clicked)


        self.lblBarB = xwidgets.XLabel()
        self.lblBarB.setFixedHeight(self.W)
        lay.addWidget(self.lblBarB, 3)
        self.lblBarB.sigClicked.connect(self.on_lbl_clicked)


        self.up_widgets()
        
    def on_lbl_clicked(self):
        self.sigSelected.emit(self.ogtGroup.group_code)

    def set_selected(self, state):
        self.selected = state
        self.up_widgets()

    def up_widgets(self):

        self.lblGroupCode.setText(self.ogtGroup.group_code)
        self.lblDataRowsCount.setText("%s" % self.ogtGroup.data_rows_count)
        self.lblErrorCount.setText("%s" % self.ogtGroup.errors_warnings_count if self.ogtGroup.errors_warnings_count else "-")

        self._up_style()

    def _up_style(self):

        bg = "#444444" if self.selected else "#999999"
        px = 2
        borders =  "border-left: %spx solid %s; " % (px, bg)
        borders += "border-right: %spx solid %s; " % (px, bg)


        #= Top bar is special if group is required
        if self.ogtGroup.group_required:
            col = "#990000" #required color always red
        else:
            col = bg
        s = "background-color: %s;" % (col)
        self.lblBarTop.setStyleSheet(s)

        #= group lbl color
        fg = "#dddddd" if self.selected else "#444444"
        gbg = "#555555" if self.selected else bg
        s = "background-color: %s;" % gbg
        s += "color: %s;" % fg
        s += borders
        self.lblGroupCode.setStyleSheet(s)

        #= Errors count
        c = "transparent" if self.ogtGroup.errors_warnings_count == 0 else CELL_COLORS.err_bg
        s = "background-color: %s;" % c
        s += borders
        self.lblErrorCount.setStyleSheet(s)

        s = "background-color: #E8E8E8;"
        s += borders
        self.lblDataRowsCount.setStyleSheet(s)


        ##  bottom
        s = "background-color: %s;" % bg
        self.lblBarB.setStyleSheet(s)


    def on_ags_info(self):
        ags4_widgets.show_group_dialog(self, group_code=self.ogtGroup.group_code)






# -*- coding: utf-8 -*-
"""
@author: Peter Morgan <pete@daffodil.uk.com>
"""

from Qt import Qt, QtGui, QtCore

import xwidgets


DATA_ROLE = Qt.UserRole + 20
SORT_NO = "sort_no"
SORT_DATE = "sort_date"


class XSortFilterProxyModel(QtCore.QSortFilterProxyModel):

    def __init__( self, parent=None, active=None, source_model=None):
        QtCore.QSortFilterProxyModel.__init__( self, parent )
        #super(QtGui.QSortFilterProxyModel, self).__init__(parent)

        self.active = active



        self.setSortCaseSensitivity(Qt.CaseInsensitive)
        self.setFilterCaseSensitivity(Qt.CaseInsensitive)
        self.setDynamicSortFilter(True)

        if source_model:
            self.setSourceModel(source_model)

    def lessThan(self, lidx, ridx):

        #
        # sm = self.sourceModel()

        #hi = sm.horizontalHeaderItem( lidx.column() ).data(DATA_ROLE).toString()
        #sort_role = sm.horizontalHeaderItem( lidx.column() ).data(DATA_ROLE).toString()

        #if sort_role.length() > 0 :
        #    if  sort_role == SORT_NO:
        #        ld, _ = sm.itemFromIndex( lidx ).text().replace(",", "").toFloat()
        #        rd, _ = sm.itemFromIndex( ridx ).text().replace(",", "").toFloat()
        #        return rd > ld

        return QtGui.QSortFilterProxyModel.lessThan(self, lidx, ridx)

    def set_active_filter(self, activ):

        self.active = activ
        self.invalidateFilter()


    def get_items_from_index(self, pidx):
        sidx = self.mapToSource(pidx)
        return self.sourceModel().get_items_from_index(sidx)

    def get_dic_from_index(self, pidx):
        sidx = self.mapToSource(pidx)
        return self.sourceModel().get_dic_from_index(sidx)


    def get_dic_items_from_index(self, pidx):
        return self.get_dic_from_index(pidx), self.get_items_from_index(pidx)


##==============================================================================================
class XStandardItemModel( QtGui.QStandardItemModel ):
    """An extended model with convenience functions, intended as a base model"""


    def __init__( self, parent=None ):
        QtGui.QStandardItemModel.__init__( self, parent )

        self.widget = None



    def set_widget(self, widget):
        self.widget = widget



    def set_header(self, col, text, align=None, tooltip=None, sort=None):
        """Sets a horizontal header

        :param col: The column index
        :type col: int
        :param label: The text label
        :type label: str`
        :param align: Text alignment
        :type align: Qt.Align
        :param tooltip: The tooltip to show on header hover
        :type tooltip: str
        :param sort: # TBA
        :type sort: int

        """
        item = xwidgets.XStandardItem()
        item.setText(text)
        if align:
            item.setTextAlignment(align)
        if tooltip:
            item.setToolTip(tooltip)
        if sort:
            item.setData(sort, DATA_ROLE)
        self.setHorizontalHeaderItem(col, item)

    def remove_rows(self):
        self.removeRows( 0, self.rowCount() )

    def make_blank_row(self):
        arr = []
        for _ in range(0, self.columnCount()):
            item = xwidgets.XStandardItem()
            arr.append(item)

        return arr

    def make_row(self, xid, col):
        arr = []
        for c in range(0, self.columnCount()):
            item = xwidgets.XStandardItem()
            if c == col:
                item.setText(str(xid))
            arr.append(item)
        self.appendRow(arr)
        return arr[0].row()



    def set_row_bg(self, ridx, color):

        for c in range(0, self.columnCount()):
            self.item(ridx, c).set_bg(color)

    def get_items_from_item(self, item):
        ridx = item.index()
        return self.get_items_from_index(ridx)

    def get_items_from_index(self, midx):
        arr = []
        for cidx in range(0, self.columnCount()):
            idx = self.index(midx.row(), cidx, midx.parent() )
            arr.append( self.itemFromIndex( idx) )

        return arr

        #idx = self.index(sidx.row(), )
        #return self.get_items_from_row( sidx.row() )

    def get_items_from_row(self, ridx):

        arr = []
        for cidx in range(0, self.columnCount()):
            arr.append( self.item(ridx, cidx) )

        return arr


    def strike_row(self, ridx, state):
        for cidx in range(0, self.columnCount()):
            font = self.item(ridx, cidx).font()
            font.setStrikeOut(state)
            self.item(ridx, cidx).setFont(font)

